/* --------
   Includes
   -------- */

#include "Yap.h"
#if defined(YAPOR) || defined(TABLING)
#include <unistd.h>
#include <signal.h>
#ifdef TABLING
#include "tab.macros.h"
#endif /* TABLING */


/* ---------------------------
   Local functions declaration
   --------------------------- */

static void receive_signals(int s);


/* ----------------
   Global functions
   ---------------- */

void init_global(int n_workers, int sch_loop, int delay_load) {
  int i;

  /* global data related to memory management */
  INIT_LOCK(GLOBAL_lock_free_area);
#ifdef YAPOR
  GLOBAL_who_locked_heap = MAX_WORKERS;
  INIT_LOCK(GLOBAL_lock_heap_access);
  INIT_LOCK(GLOBAL_lock_bm_idle_workers);
  INIT_LOCK(GLOBAL_lock_bm_root_cp_workers);
  INIT_LOCK(GLOBAL_lock_bm_invisible_workers);
  INIT_LOCK(GLOBAL_lock_bm_requestable_workers);
  INIT_LOCK(GLOBAL_lock_bm_finished_workers);
  INIT_LOCK(GLOBAL_lock_free_or_frame);
  INIT_LOCK(GLOBAL_lock_free_answer_frame);
  INIT_LOCK(GLOBAL_lock_free_solution_frame);
#endif /* YAPOR */
#ifdef TABLING
  INIT_LOCK(GLOBAL_lock_free_table_entry);
  INIT_LOCK(GLOBAL_lock_free_subgoal_trie_node);
  INIT_LOCK(GLOBAL_lock_free_subgoal_frame);
  INIT_LOCK(GLOBAL_lock_free_answer_trie_node);
  INIT_LOCK(GLOBAL_lock_free_dependency_frame);
#endif /* TABLING */

#ifdef YAPOR
  /* global static data */
  number_workers = n_workers;
  worker_pid(0) = getpid();
  for (i = 1; i < number_workers; i++) worker_pid(i) = 0;
  SCHEDULER_LOOP = sch_loop;
  DELAYED_RELEASE_LOAD = delay_load;

  /* global data related to or-parallelism */
  BITMAP_clear(GLOBAL_bm_present_workers);
  for (i = 0; i < number_workers; i++) 
    BITMAP_insert(GLOBAL_bm_present_workers, i);
  BITMAP_copy(GLOBAL_bm_idle_workers, GLOBAL_bm_present_workers);
  BITMAP_clear(GLOBAL_bm_root_cp_workers);
  BITMAP_copy(GLOBAL_bm_finished_workers, GLOBAL_bm_present_workers);
  BITMAP_clear(GLOBAL_bm_invisible_workers);
  BITMAP_clear(GLOBAL_bm_requestable_workers);
  if (number_workers == 1)
    GLOBAL_execution_mode = SEQUENTIAL_MODE;
  else
    GLOBAL_execution_mode = PARALLEL_MODE;
  GLOBAL_parallel_mode = SUSPENDED_MODE;
#endif /* YAPOR */

#ifdef TABLING
  /* global data related to tabling */
  for (i = 0; i < MAX_TABLE_VARS; i++)
    GLOBAL_table_var_enumerator(i) = (CELL) & GLOBAL_table_var_enumerator(i);
#endif /* TABLING */
  return;
}


void init_local(void) {
#ifdef YAPOR
  /* local data related to or-parallelism */
  LOCAL = REMOTE[worker_id];
  LOCAL_top_cp = (choiceptr) LocalBase;
  LOCAL_top_or_fr = GLOBAL_root_or_fr;
  LOCAL_load = - DELAYED_RELEASE_LOAD;
  LOCAL_request = MAX_WORKERS;
  INIT_LOCK(LOCAL_lock_signals);
  LOCAL_cut_cp = NULL;
#endif /* YAPOR */
  INIT_LOCK(LOCAL_lock);
#ifdef TABLING
  /* local data related to tabling */
#ifdef YAPOR
  LOCAL_top_consumer_cp = B_BASE;
  LOCAL_top_collect_susp_fr = NULL;
#endif /* YAPOR */
  LOCAL_top_dep_fr = GLOBAL_root_dep_fr; 
  LOCAL_top_sg_fr = NULL; 
  H_FZ = H_BASE;
  B_FZ = B_BASE;
  TR_FZ = TR_BASE;
#else
#ifdef SBA /* vsc_sba */
  H_FZ = H_BASE;
  B_FZ = B_BASE;
  TR_FZ = TR_BASE;
#endif
#endif /* TABLING */
  return;
}


void make_root_frames(void) {
#ifdef YAPOR
  or_fr_ptr or_fr;
#endif  /* YAPOR */
#ifdef TABLING
  dep_fr_ptr dep_fr;
#endif /* TABLING */

#ifdef YAPOR
  /* root or frame */
  ALLOC_OR_FRAME(or_fr);   
  INIT_LOCK(OrFr_lock(or_fr));
  OrFr_alternative(or_fr) = NULL;
  BITMAP_copy(OrFr_members(or_fr), GLOBAL_bm_present_workers);
  OrFr_node(or_fr) = NULL;
  OrFr_nearest_livenode(or_fr) = NULL;
  OrFr_depth(or_fr) = 0;
  OrFr_nearest_leftnode(or_fr) = NULL;
  OrFr_pend_cut_cp(or_fr) = NULL;
  OrFr_solutions(or_fr) = NULL;
#ifdef TABLING
  BITMAP_clear(OrFr_hidden_members(or_fr));
  OrFr_next_on_branch(or_fr) = NULL;
  OrFr_suspensions(or_fr) = NULL;
#endif /* TABLING */
  OrFr_next(or_fr) = NULL;
  GLOBAL_root_or_fr = or_fr;
#endif /* YAPOR */
#ifdef TABLING
  /* root dependency frame */
  ALLOC_DEPENDENCY_FRAME(dep_fr);
  INIT_LOCK(DepFr_lock(dep_fr));
  DepFr_cons_cp(dep_fr) = CONS_CP(B_BASE);
  DepFr_leader_cp(dep_fr) = NULL ;
  DepFr_sg_fr(dep_fr) = NULL;
  DepFr_last_ans(dep_fr) = NULL;
  DepFr_h_fz(dep_fr) = H_BASE;
  DepFr_b_fz(dep_fr) = B_BASE;
  DepFr_tr_fz(dep_fr) = TR_BASE;
#ifdef YAPOR
  DepFr_top_chain_cp(dep_fr) = NULL;
#endif /* YAPOR */
  DepFr_next(dep_fr) = NULL;
  GLOBAL_root_dep_fr = dep_fr;
#endif /* TABLING */
}


void init_signals(void) {

return;

  signal(SIGSEGV, receive_signals);
  signal(SIGABRT, receive_signals);
  signal(SIGFPE, receive_signals);
  signal(SIGHUP, receive_signals);
  signal(SIGINT, receive_signals);
  signal(SIGKILL, receive_signals);
  signal(SIGQUIT, receive_signals);
  signal(SIGSTOP, receive_signals);
  signal(SIGTERM, receive_signals);
  signal(SIGTSTP, receive_signals);
  return;
}


#ifdef YAPOR
void init_workers(int fd) {
  int proc;
  for (proc = 1; proc < number_workers; proc++) {
    int son;
    son = fork();
    if (son == -1) abort_optyap("fork error in function init_workers");
    if (son == 0) { 
      /* new worker */
      worker_id = proc;
      remap_memory(fd);
      break;
    }
    else worker_pid(proc) = son;
  }
}
#endif /* YAPOR */


/* ---------------
   Local functions
   --------------- */

static void receive_signals(int s) {
  abort_optyap("receiving signal number %d", s);
}
#endif /* YAPOR || TABLING */
