/* ---------------------
   Atomic lock for X86
   --------------------- */

#ifdef i386
#define swap(reg,adr)					     \
({							     \
   char _ret;						     \
   asm volatile ("xchgb %0,%1"				     \
	: "=q" (_ret), "=m" (*(adr))	/* Output %0,%1 */   \
	: "m"  (*(adr)), "0"  (reg));	/* Input (%2),%0 */  \
   _ret;						     \
})
#define TRY_LOCK(p)   (swap(1,(p))==0)
#endif linux


/* ---------------------
   Atomic lock for SPARC
   --------------------- */

#ifdef sparc
#define swap_il(adr,reg)				     \
({ int _ret;						     \
   asm volatile ("swap %1,%0"				     \
	: "=r" (_ret), "=m" (*(adr))	/* Output %0,%1 */   \
	: "m"  (*(adr)), "0" (reg));	/* Input (%2),%0 */  \
   _ret;						     \
})
#define TRY_LOCK(p)   (swap_il((p),1)==0)
#endif sparc


/* ------------------
   Locking operations
   ------------------ */

#define IS_LOCKED(p)   ((p) == 1)
#define INIT_LOCK(p)   ((p) = 0)
#define UNLOCK(p)      ((p) = 0)
#define LOCK(p)        do {                                     \
                            if (TRY_LOCK(&(p))) break;          \
			    SIMICS_SERVICE(START_COUNTER, 11);  \
		            while (IS_LOCKED(p)) continue;      \
			    SIMICS_SERVICE(STOP_COUNTER, 11);   \
		          } while (1)


/* ---------------------------
   Alloc and free data structs
   --------------------------- */

#ifdef YAPOR
#define ALLOC_OR_FRAME(fr)                                       \
        LOCK(GLOBAL_lock_free_or_frame);                         \
        if (GLOBAL_free_or_frame == NULL) alloc_or_frames();     \
        fr = GLOBAL_free_or_frame;                               \
        GLOBAL_free_or_frame = OrFr_next(GLOBAL_free_or_frame);  \
        UNLOCK(GLOBAL_lock_free_or_frame)

#define FREE_OR_FRAME(fr)                      \
        LOCK(GLOBAL_lock_free_or_frame);       \
        OrFr_next(fr) = GLOBAL_free_or_frame;  \
	GLOBAL_free_or_frame = fr;             \
	UNLOCK(GLOBAL_lock_free_or_frame)

#define ALLOC_ANSWER_FRAME(fr)                                            \
        LOCK(GLOBAL_lock_free_answer_frame);                              \
        if (GLOBAL_free_answer_frame == NULL) alloc_answer_frames();      \
        fr = GLOBAL_free_answer_frame;                                    \
        GLOBAL_free_answer_frame = AnsFr_next(GLOBAL_free_answer_frame);  \
        UNLOCK(GLOBAL_lock_free_answer_frame)

#define FREE_ANSWER_FRAME(fr)                       \
        LOCK(GLOBAL_lock_free_answer_frame);        \
        AnsFr_next(fr) = GLOBAL_free_answer_frame;  \
	GLOBAL_free_answer_frame = fr;              \
	UNLOCK(GLOBAL_lock_free_answer_frame)

#define ALLOC_SOLUTION_FRAME(fr)                                              \
        LOCK(GLOBAL_lock_free_solution_frame);                                \
        if (GLOBAL_free_solution_frame == NULL) alloc_solution_frames();      \
        fr = GLOBAL_free_solution_frame;                                      \
        GLOBAL_free_solution_frame = SolFr_next(GLOBAL_free_solution_frame);  \
        UNLOCK(GLOBAL_lock_free_solution_frame)

#define FREE_SOLUTION_FRAME(fr)                       \
	LOCK(GLOBAL_lock_free_solution_frame);        \
	SolFr_next(fr) = GLOBAL_free_solution_frame;  \
        GLOBAL_free_solution_frame = fr;              \
        UNLOCK(GLOBAL_lock_free_solution_frame)
#endif /* YAPOR */

#ifdef TABLING
#define ALLOC_TABLE_ENTRY(te)                                            \
        LOCK(GLOBAL_lock_free_table_entry);                              \
        if (GLOBAL_free_table_entry == NULL) alloc_table_entries();      \
        te = GLOBAL_free_table_entry;                                    \
        GLOBAL_free_table_entry = TabEnt_next(GLOBAL_free_table_entry);  \
        UNLOCK(GLOBAL_lock_free_table_entry)

#define FREE_TABLE_ENTRY(te)                        \
	LOCK(GLOBAL_lock_free_table_entry);         \
	TabEnt_next(te) = GLOBAL_free_table_entry;  \
        GLOBAL_free_table_entry = te;               \
        UNLOCK(GLOBAL_lock_free_table_entry)

#define ALLOC_SUBGOAL_TRIE_NODE(tr)                                                  \
        LOCK(GLOBAL_lock_free_subgoal_trie_node);                                    \
        if (GLOBAL_free_subgoal_trie_node == NULL) alloc_subgoal_trie_nodes();       \
        tr = GLOBAL_free_subgoal_trie_node;                                          \
        GLOBAL_free_subgoal_trie_node = TrNode_next(GLOBAL_free_subgoal_trie_node);  \
        UNLOCK(GLOBAL_lock_free_subgoal_trie_node)

#define FREE_SUBGOAL_TRIE_NODE(tr)                        \
	LOCK(GLOBAL_lock_free_subgoal_trie_node);         \
	TrNode_next(tr) = GLOBAL_free_subgoal_trie_node;  \
        GLOBAL_free_subgoal_trie_node = tr;               \
        UNLOCK(GLOBAL_lock_free_subgoal_trie_node)

#define ALLOC_SUBGOAL_FRAME(fr)                                            \
        LOCK(GLOBAL_lock_free_subgoal_frame);                              \
        if (GLOBAL_free_subgoal_frame == NULL) alloc_subgoal_frames();     \
        fr = GLOBAL_free_subgoal_frame;                                    \
        GLOBAL_free_subgoal_frame = SgFr_next(GLOBAL_free_subgoal_frame);  \
        UNLOCK(GLOBAL_lock_free_subgoal_frame)

#define FREE_SUBGOAL_FRAME(fr)                      \
	LOCK(GLOBAL_lock_free_subgoal_frame);       \
	SgFr_next(fr) = GLOBAL_free_subgoal_frame;  \
        GLOBAL_free_subgoal_frame = fr;             \
        UNLOCK(GLOBAL_lock_free_subgoal_frame)

#define ALLOC_ANSWER_TRIE_NODE(tr)                                                 \
        LOCK(GLOBAL_lock_free_answer_trie_node);                                   \
        if (GLOBAL_free_answer_trie_node == NULL) alloc_answer_trie_nodes();       \
        tr = GLOBAL_free_answer_trie_node;                                         \
        GLOBAL_free_answer_trie_node = TrNode_next(GLOBAL_free_answer_trie_node);  \
        UNLOCK(GLOBAL_lock_free_answer_trie_node)

#define FREE_ANSWER_TRIE_NODE(tr)                        \
	LOCK(GLOBAL_lock_free_answer_trie_node);         \
	TrNode_next(tr) = GLOBAL_free_answer_trie_node;  \
        GLOBAL_free_answer_trie_node = tr;               \
        UNLOCK(GLOBAL_lock_free_answer_trie_node)

#define ALLOC_DEPENDENCY_FRAME(fr)                                                \
        LOCK(GLOBAL_lock_free_dependency_frame);                                  \
        if (GLOBAL_free_dependency_frame == NULL) alloc_dependency_frames();      \
        fr = GLOBAL_free_dependency_frame;                                        \
        GLOBAL_free_dependency_frame = DepFr_next(GLOBAL_free_dependency_frame);  \
        UNLOCK(GLOBAL_lock_free_dependency_frame)

#define FREE_DEPENDENCY_FRAME(fr)                       \
	LOCK(GLOBAL_lock_free_dependency_frame);        \
	DepFr_next(fr) = GLOBAL_free_dependency_frame;  \
        GLOBAL_free_dependency_frame = fr;              \
        UNLOCK(GLOBAL_lock_free_dependency_frame)

#ifdef YAPOR
#define ALLOC_SUSPENSION_FRAME(fr)                                                 \
        LOCK(GLOBAL_lock_free_suspension_frame);                                   \
        if (GLOBAL_free_suspension_frame == NULL) alloc_suspension_frames();       \
        fr = GLOBAL_free_suspension_frame;                                         \
        GLOBAL_free_suspension_frame = SuspFr_next(GLOBAL_free_suspension_frame);  \
        UNLOCK(GLOBAL_lock_free_suspension_frame)

#define FREE_SUSPENSION_FRAME(fr)                        \
	LOCK(GLOBAL_lock_free_suspension_frame);         \
	SuspFr_next(fr) = GLOBAL_free_suspension_frame;  \
        GLOBAL_free_suspension_frame = fr;               \
        UNLOCK(GLOBAL_lock_free_suspension_frame)
#endif /* YAPOR */
#endif /* TABLING */


/* ---------------------------
   Bitmap tests and operations
   --------------------------- */

#define BITMAP_empty(b)		        (b == 0)
#define BITMAP_member(b,n)	        ((b&(1<<(n))) != 0)
#define BITMAP_alone(b,n)	        (b == (1<<(n)))
#define BITMAP_subset(b1,b2)	        ((b1&b2) == b2)
#define BITMAP_intersect(b1,b2)	        ((b1&b2) != 0)
#define BITMAP_intersect3(b1,b2,b3)     ((b1&b2&b3) != 0)
#define BITMAP_same(b1,b2)	        (b1 == b2)

#define BITMAP_clear(b)		        (b = 0)
#define BITMAP_or(b1,b2)	        (b1 |= b2)
#define BITMAP_and(b1,b2)	        (b1 &= b2)
#define BITMAP_minus(b1,b2)             (b1 &= ~b2)
#define BITMAP_insert(b,n)	        (b |= (1<<(n)))
#define BITMAP_delete(b,n)	        (b &= (~(1<<(n))))
#define BITMAP_copy(b1,b2)	        (b1 = b2)
#define BITMAP_invert(b)	        (b = ~b)
#define BITMAP_inverse(b1,b2)	        (b1 = ~b2)
#define BITMAP_union(b1,b2,b3)	        (b1 = (b2|b3))
#define BITMAP_intersection(b1,b2,b3)   (b1 = (b2&b3))
#define BITMAP_difference(b1,b2,b3)	(b1 = (b2&(~b3)))


/* -------------------
   Miscelaneous macros
   ------------------- */

#if defined(TABLING) || (defined(YAPOR) && defined(SBA)) /* vsc_sba */
#define H_BASE   ((CELL *) GlobalBase)
#define B_BASE   ((choiceptr) LocalBase)
#define TR_BASE  ((tr_fr_ptr) TrailBase)
#endif /* TABLING  || SBA */

#ifdef YAPOR
#define SEQUENTIAL_MODE   0
#define PARALLEL_MODE     1
#define EXECUTING_MODE    0
#define SUSPENDED_MODE    1
#define YAMOP_LTT  u.ld.ltt
#define YAMOP_CUT  u.ld.cut
#define YAMOP_SEQ  u.ld.seq
#endif /* YAPOR */

/*  
**  type 0: general messages
**  type 1: optyap messages
**  type 2: table instructions messages
**  type 3: trie instructions messages
**  type 4: or-parallel messages (not used yet) 
*/
#define optyap_message_type0(message, args...) optyap_print_message(message, ##args)
#define optyap_message_type1(message, args...)
#define optyap_message_type2(message, args...)
#define optyap_message_type3(message, args...)
#define optyap_message_type4(message, args...) optyap_print_message(message, ##args)
#define optyap_debug_message(message, args...) optyap_print_message(message, ##args)


/* -------------
   SimICS macros
   ------------- */

/*  Counter  0. Total time
**  Counter  1. Prolog
**  Counter  2. P Share
**  Counter  3. Scheduler
**  Counter  4. Cut request
**  Counter  5. End operations
**  Counter  6. Cut shared (Counter 1 or 3 or 5)
**  Counter  7. Number of requests to share work (Counter 3)
**  Counter  8. Number of refused requests (Counter 3)
**  Counter  9. Number of tasks (Counter 1)
**  Counter 10. Number of calls (Counter 1)
**  Counter 11. Number of failed TRY_LOCK's  */

#define START_COUNTER            1
#define STOP_COUNTER             2

#define TOTAL_TIME               0
#define PROLOG                   1
#define SHARE                    2
#define SCHEDULER                3
#define CUT_REQUEST              4
#define END_OPERATIONS           5
#define CUT_SHARED               6
#define ONE_MORE_REQUEST         7
#define ONE_MORE_REFUSED_REQUEST 8
#define ONE_MORE_TASK            9
#define ONE_MORE_CALL            10

#ifdef SIMICS
#define SIMICS_ATOMIC_SERVICE(COUNTER)           \
        SIMICS_SERVICE(START_COUNTER, COUNTER);  \
        SIMICS_SERVICE(STOP_COUNTER, COUNTER)
#define SIMICS_SERVICE(COMMAND, COUNTER)         \
        do {                                     \
          asm volatile ("sethi %0, %%g0" :       \
          /* no outputs */ :                     \
          "g" ((COMMAND << 16) | COUNTER));      \
	} while(0)
#else
#define SIMICS_ATOMIC_SERVICE(COUNTER)
#define SIMICS_SERVICE(COMMAND, COUNTER)
#endif SIMICS
