/* -------------------------------------
   Includes, defines and local variables
   ------------------------------------- */

#include "Yap.h"
#ifdef YAPOR
#include <sys/time.h>
#include <stdio.h>

#define TIME_RESOLUTION 1000000


/* ----------------
   Global functions
   ---------------- */

realtime current_time(void) {
  /* to get time as Yap */
  /*
  double now, interval;
  cputime_interval(&now, &interval);
  return ((realtime)now);
  */
  struct timeval tempo;
  gettimeofday(&tempo, NULL);
  return ((realtime)tempo.tv_sec + (realtime)tempo.tv_usec / TIME_RESOLUTION);
}


int p_performance(void) {
  int i;
  realtime one_worker_execution_time = 0;
  Term t;

  t = Deref(ARG1);
  if (IsIntTerm(t)) one_worker_execution_time = IntOfTerm(t);
  if (IsFloatTerm(t)) one_worker_execution_time = FloatOfTerm(t);

  fprintf(stdout, "[\n  Best execution times:\n");
  for (i = 1; i <= GLOBAL_number_goals; i++) {
    fprintf(stdout, "    %d. time: %f seconds", i, GLOBAL_best_times(i));  
    if (one_worker_execution_time != 0)
      fprintf(stdout, " --> speedup %f (%6.2f %% )\n",
              one_worker_execution_time / GLOBAL_best_times(i),
              one_worker_execution_time / GLOBAL_best_times(i) / number_workers * 100 );
    else fprintf(stdout, "\n");
  }

  fprintf(stdout, "  Average             : %f seconds",
          GLOBAL_best_times(0) / GLOBAL_number_goals);
  if (one_worker_execution_time != 0)
    fprintf(stdout, " --> speedup %f (%6.2f %% )",
            one_worker_execution_time * GLOBAL_number_goals / GLOBAL_best_times(0),
            one_worker_execution_time * GLOBAL_number_goals / GLOBAL_best_times(0) / number_workers * 100 );

  if (GLOBAL_number_goals >= 3) {
    fprintf(stdout, "\n  Average (best three): %f seconds",
            (GLOBAL_best_times(1) + GLOBAL_best_times(2) + GLOBAL_best_times(3)) / 3);
    if (one_worker_execution_time != 0)
      fprintf(stdout, " --> speedup %f (%6.2f %% ) ]\n\n",
              one_worker_execution_time * 3 / (GLOBAL_best_times(1) + GLOBAL_best_times(2) + GLOBAL_best_times(3)),
              one_worker_execution_time * 3 / (GLOBAL_best_times(1) + GLOBAL_best_times(2) + GLOBAL_best_times(3)) / number_workers * 100 );
    else fprintf(stdout, "\n]\n\n");
  } else fprintf(stdout, "\n]\n\n");

  return (TRUE);
}
#endif /* YAPOR */
