/* --------------------
   Typedefs and defines
   -------------------- */

typedef volatile int lockvar;
typedef unsigned long bitmap;
typedef double realtime;

#define MAX_WORKERS (sizeof(bitmap) * 8)
#define MAX_LENGTH_ANSWER 500
#define MAX_DEPTH 200
#define MAX_BEST_TIMES 21
#define MAX_TABLE_VARS 512



/* -------------------
   struct global_locks 
   ------------------- */

struct global_locks{
  lockvar free_area;
#ifdef YAPOR
  lockvar who_locked_heap;  /* used in CriticalSection's macros */
  lockvar heap_access;      /* used in CriticalSection's macros */

  lockvar bitmap_idle_workers;
  lockvar bitmap_root_cp_workers;
  lockvar bitmap_invisible_workers;
  lockvar bitmap_requestable_workers;
  lockvar bitmap_finished_workers;

  lockvar free_or_frame;
  lockvar free_answer_frame;
  lockvar free_solution_frame;
#endif /* YAPOR */
#ifdef TABLING
  lockvar free_table_entry;
  lockvar free_subgoal_trie_node;
  lockvar free_subgoal_frame;
  lockvar free_answer_trie_node;
  lockvar free_dependency_frame;
#ifdef YAPOR
  lockvar free_suspension_frame;
#endif /* YAPOR */
#endif /* TABLING */
};



/* ------------------
   struct global_data
   ------------------ */

struct global_data{
#ifdef YAPOR
  /* global data inherted from Yap*/
  Atom HashChain[MaxHash];
  ADDR HeapTop;
  Int HeapMax;
  Int HeapUsed;
  void *FreeBlocks;
#endif /* YAPOR */

  /* global data related to memory management */
  void *BaseWorkArea;
  void *TopWorkArea;
  void *FramesArea;
#if defined(YAPOR) && defined(TABLING)
  void *SuspensionArea;
#endif /* YAPOR && TABLING */
  struct global_locks locks;
#ifdef YAPOR
  struct or_frame *next_free_or_frame;
  struct answer_frame *next_free_answer_frame;
  struct solution_frame *next_free_solution_frame;
#endif /* YAPOR */
#ifdef TABLING
  struct table_entry *next_free_table_entry;
  struct subgoal_trie_node *next_free_subgoal_trie_node;
  struct subgoal_frame *next_free_subgoal_frame;
  struct answer_trie_node *next_free_answer_trie_node;
  struct dependency_frame *next_free_dependency_frame;
#ifdef YAPOR
  struct suspension_frame *next_free_suspension_frame;
#endif /* YAPOR */
#endif /* TABLING */

#ifdef YAPOR
  /* global static data */
  int number_workers;
  int worker_pid[MAX_WORKERS];
  int scheduler_loop;
  int delayed_release_load;

  /* global data related to performance */
  realtime execution_time;
  realtime best_execution_times[MAX_BEST_TIMES];
  int number_of_executed_goals;

  /* global data related to or-parallelism */
  choiceptr root_choice_point;
  struct or_frame *root_or_frame;
  volatile struct {
    int alt_ltt;
    int alt_cut;
  } branch[MAX_WORKERS][MAX_DEPTH];
  bitmap present_workers;
  volatile bitmap idle_workers;
  volatile bitmap root_cp_workers;
  volatile bitmap invisible_workers;
  volatile bitmap requestable_workers;
  volatile bitmap finished_workers;
  volatile char execution_mode;  /* SEQUENTIAL_MODE / PARALLEL_MODE */
  volatile char parallel_mode;   /* EXECUTING_MODE / SUSPENDED_MODE */
  volatile int answers;
#endif /* YAPOR */

#ifdef TABLING
  /* global data related to tabling */
  struct dependency_frame *root_dependency_frame;
  CELL table_var_enumerator[MAX_TABLE_VARS];
#endif /* TABLING */
};

extern struct global_data *GLOBAL;

#ifdef YAPOR
#define HashChain                             (GLOBAL->HashChain)
#define HeapTop                               (GLOBAL->HeapTop)
#define HeapMax                               (GLOBAL->HeapMax)
#define HeapUsed                              (GLOBAL->HeapUsed)
#define FreeBlocks                            (GLOBAL->FreeBlocks)
#endif /* YAPOR */
#define BaseWorkArea                          (GLOBAL->BaseWorkArea)
#define TopWorkArea                           (GLOBAL->TopWorkArea)
#define FramesArea                            (GLOBAL->FramesArea)
#define SuspensionArea                        (GLOBAL->SuspensionArea)
#define GLOBAL_lock_free_area                 (GLOBAL->locks.free_area)
#define GLOBAL_who_locked_heap                (GLOBAL->locks.who_locked_heap)
#define GLOBAL_lock_heap_access               (GLOBAL->locks.heap_access)
#define GLOBAL_lock_bm_idle_workers           (GLOBAL->locks.bitmap_idle_workers)
#define GLOBAL_lock_bm_root_cp_workers        (GLOBAL->locks.bitmap_root_cp_workers)
#define GLOBAL_lock_bm_invisible_workers      (GLOBAL->locks.bitmap_invisible_workers)
#define GLOBAL_lock_bm_requestable_workers    (GLOBAL->locks.bitmap_requestable_workers)
#define GLOBAL_lock_bm_finished_workers       (GLOBAL->locks.bitmap_finished_workers)
#define GLOBAL_lock_free_or_frame             (GLOBAL->locks.free_or_frame)
#define GLOBAL_lock_free_answer_frame         (GLOBAL->locks.free_answer_frame)
#define GLOBAL_lock_free_solution_frame       (GLOBAL->locks.free_solution_frame)
#define GLOBAL_lock_free_table_entry          (GLOBAL->locks.free_table_entry)
#define GLOBAL_lock_free_subgoal_trie_node    (GLOBAL->locks.free_subgoal_trie_node)
#define GLOBAL_lock_free_subgoal_frame        (GLOBAL->locks.free_subgoal_frame)
#define GLOBAL_lock_free_answer_trie_node     (GLOBAL->locks.free_answer_trie_node)
#define GLOBAL_lock_free_dependency_frame     (GLOBAL->locks.free_dependency_frame)
#define GLOBAL_lock_free_suspension_frame     (GLOBAL->locks.free_suspension_frame)
#define GLOBAL_free_or_frame                  (GLOBAL->next_free_or_frame)
#define GLOBAL_free_answer_frame              (GLOBAL->next_free_answer_frame)
#define GLOBAL_free_solution_frame            (GLOBAL->next_free_solution_frame)
#define GLOBAL_free_table_entry               (GLOBAL->next_free_table_entry)
#define GLOBAL_free_subgoal_trie_node         (GLOBAL->next_free_subgoal_trie_node)
#define GLOBAL_free_subgoal_frame             (GLOBAL->next_free_subgoal_frame)
#define GLOBAL_free_answer_trie_node          (GLOBAL->next_free_answer_trie_node)
#define GLOBAL_free_dependency_frame          (GLOBAL->next_free_dependency_frame)
#define GLOBAL_free_suspension_frame          (GLOBAL->next_free_suspension_frame)
#define number_workers                        (GLOBAL->number_workers)
#define worker_pid(w)                         (GLOBAL->worker_pid[w])
#define SCHEDULER_LOOP                        (GLOBAL->scheduler_loop)
#define DELAYED_RELEASE_LOAD                  (GLOBAL->delayed_release_load)
#define GLOBAL_execution_time                 (GLOBAL->execution_time)
#define GLOBAL_best_times(t)                  (GLOBAL->best_execution_times[t])
#define GLOBAL_number_goals                   (GLOBAL->number_of_executed_goals)
#define GLOBAL_root_cp                        (GLOBAL->root_choice_point)
#define GLOBAL_root_or_fr                     (GLOBAL->root_or_frame)
#define GLOBAL_branch_ltt(w,d)                (GLOBAL->branch[w][d].alt_ltt)
#define GLOBAL_branch_cut(w,d)                (GLOBAL->branch[w][d].alt_cut)
#define GLOBAL_bm_present_workers             (GLOBAL->present_workers)
#define GLOBAL_bm_idle_workers                (GLOBAL->idle_workers)
#define GLOBAL_bm_root_cp_workers             (GLOBAL->root_cp_workers)
#define GLOBAL_bm_invisible_workers           (GLOBAL->invisible_workers)
#define GLOBAL_bm_requestable_workers         (GLOBAL->requestable_workers)
#define GLOBAL_bm_finished_workers            (GLOBAL->finished_workers)
#define GLOBAL_execution_mode                 (GLOBAL->execution_mode)
#define GLOBAL_parallel_mode                  (GLOBAL->parallel_mode)
#define GLOBAL_answers                        (GLOBAL->answers)
#define GLOBAL_root_dep_fr                    (GLOBAL->root_dependency_frame)
#define GLOBAL_table_var_enumerator(i)        (GLOBAL->table_var_enumerator[i])



/* --------------------
   struct local_signals
   -------------------- */

#ifdef YAPOR
struct local_signals{
  volatile enum {
    no_reply = 0,
    accepted = 1,
    local_ok = 2,
    install  = 3,
    ready    = 4,
    refused  = 5 
  } reply;
  volatile enum {
    Q_idle = 0,
    trail  = 1,
    global = 2,
    local  = 3,
    P_idle = 4
  } P_fase, Q_fase;
  lockvar lock;
};
#endif /* YAPOR */



/* -----------------
   struct local_data
   ----------------- */

struct local_data{
  lockvar lock;

#ifdef YAPOR
  /* local data related to or-parallelism */
  choiceptr top_choice_point;     /* top shared choice point */
  struct or_frame *top_or_frame;  /* top or frame */
  volatile int load;
  volatile int request_from_worker;
  struct local_signals sharing_signals;
  volatile struct {
    CELL start;
    CELL end;
  } global_copy, local_copy, trail_copy;
  choiceptr cut_choice_point;
#endif /* YAPOR */

#ifdef TABLING
  /* local data related to tabling */
#ifdef YAPOR
  choiceptr top_consumer_choice_point;  /* top shared consumer choice point */
  struct suspension_frame *top_collected_suspension_frame;
#endif /* YAPOR */
  struct dependency_frame *top_dependency_frame;
  struct subgoal_frame *top_subgoal_frame;
  CELL *h_freeze_register;
  choiceptr b_freeze_register;
  tr_fr_ptr tr_freeze_register;
#endif /* TABLING */
};

extern struct local_data *LOCAL;
#ifdef YAPOR
extern struct local_data *REMOTE[MAX_WORKERS];
#endif /* YAPOR */


#define LOCAL_lock                     (LOCAL->lock)
#define LOCAL_top_cp                   (LOCAL->top_choice_point)
#define LOCAL_top_or_fr                (LOCAL->top_or_frame)
#define LOCAL_load                     (LOCAL->load)
#define LOCAL_request                  (LOCAL->request_from_worker)
#define LOCAL_reply_signal             (LOCAL->sharing_signals.reply)
#define LOCAL_p_fase_signal            (LOCAL->sharing_signals.P_fase)
#define LOCAL_q_fase_signal            (LOCAL->sharing_signals.Q_fase)
#define LOCAL_lock_signals             (LOCAL->sharing_signals.lock)
#define LOCAL_start_global_copy        (LOCAL->global_copy.start)
#define LOCAL_end_global_copy          (LOCAL->global_copy.end)
#define LOCAL_start_local_copy         (LOCAL->local_copy.start)
#define LOCAL_end_local_copy           (LOCAL->local_copy.end)
#define LOCAL_start_trail_copy         (LOCAL->trail_copy.start)
#define LOCAL_end_trail_copy           (LOCAL->trail_copy.end)
#define LOCAL_cut_cp                   (LOCAL->cut_choice_point)
#define LOCAL_top_consumer_cp          (LOCAL->top_consumer_choice_point)
#define LOCAL_top_collect_susp_fr      (LOCAL->top_collected_suspension_frame)
#define LOCAL_top_dep_fr               (LOCAL->top_dependency_frame)
#define LOCAL_top_sg_fr                (LOCAL->top_subgoal_frame)
#if !defined(YAPOR) || !defined(SBA)
#define H_FZ                           (LOCAL->h_freeze_register)
#define B_FZ                           (LOCAL->b_freeze_register)
#define TR_FZ                          (LOCAL->tr_freeze_register)
#endif /* !YAPOR || !SBA */

#define REMOTE_lock(i)                 (REMOTE[i]->lock)
#define REMOTE_top_cp(i)               (REMOTE[i]->top_choice_point)
#define REMOTE_top_or_fr(i)            (REMOTE[i]->top_or_frame)
#define REMOTE_load(i)                 (REMOTE[i]->load)
#define REMOTE_request(i)              (REMOTE[i]->request_from_worker)
#define REMOTE_reply_signal(i)         (REMOTE[i]->sharing_signals.reply)
#define REMOTE_p_fase_signal(i)        (REMOTE[i]->sharing_signals.P_fase)
#define REMOTE_q_fase_signal(i)        (REMOTE[i]->sharing_signals.Q_fase)
#define REMOTE_lock_signals(i)         (REMOTE[i]->sharing_signals.lock)
#define REMOTE_start_global_copy(i)    (REMOTE[i]->global_copy.start)
#define REMOTE_end_global_copy(i)      (REMOTE[i]->global_copy.end)
#define REMOTE_start_local_copy(i)     (REMOTE[i]->local_copy.start)
#define REMOTE_end_local_copy(i)       (REMOTE[i]->local_copy.end)
#define REMOTE_start_trail_copy(i)     (REMOTE[i]->trail_copy.start)
#define REMOTE_end_trail_copy(i)       (REMOTE[i]->trail_copy.end)
#define REMOTE_cut_cp(i)               (REMOTE[i]->cut_choice_point)
#define REMOTE_top_consumer_cp(i)      (REMOTE[i]->top_consumer_choice_point)
#define REMOTE_top_collect_susp_fr(i)  (REMOTE[i]->top_collected_suspension_frame)
#define REMOTE_top_dep_fr(i)           (REMOTE[i]->top_dependency_frame)
#define REMOTE_top_sg_fr(i)            (REMOTE[i]->top_subgoal_frame)



#ifdef YAPOR
#include "or.structs.h"
#endif /* YAPOR */



#ifdef TABLING
#include "tab.structs.h"
#endif  /* TABLING */
