/* --------
   Includes
   -------- */

#include "Yap.h"
#ifdef YAPOR
#include "or.macros.h"


/* ----------------
   Global functions
   ---------------- */

int no_workers_left_pending_cut(void) {
  int i;
  for (i = 0; i < number_workers; i++)
    if (BITMAP_member(OrFr_members(LOCAL_top_or_fr), i) && 
       GLOBAL_branch_ltt(i, OrFr_depth(LOCAL_top_or_fr)) >= OrFr_pend_cut_ltt(LOCAL_top_or_fr))
      return 0;
  return 1;
}


void cut_shared(choiceptr cut_cp, sol_fr_ptr aux_solution) {
  int i, cut_more = 1;
  choiceptr leftmost_cp;
  bitmap aux_bitmap;
  or_fr_ptr aux_frame;

  leftmost_cp = CUT_leftmost_until();
  if(leftmost_cp < cut_cp) goto not_leftmost;

  /* performing a cut in the shared region, being leftmost */
  BITMAP_clear(aux_bitmap);
  BITMAP_insert(aux_bitmap, worker_id);
  while (LOCAL_top_cp != cut_cp) {
    LOCK_OR_FRAME(LOCAL_top_or_fr);
    OrFr_alternative(LOCAL_top_or_fr) = NULL;
    BITMAP_difference(aux_bitmap, OrFr_members(LOCAL_top_or_fr), aux_bitmap);
    for (i = 0; i < number_workers; i++)
      if (BITMAP_member(aux_bitmap, i)) { 
        CUT_make_cut_request(i, cut_cp); 
      } 
    aux_solution = CUT_put_not_cut_solutions_together(aux_solution);
    BITMAP_copy(aux_bitmap, OrFr_members(LOCAL_top_or_fr));
    BITMAP_delete(OrFr_members(LOCAL_top_or_fr), worker_id);
    if (BITMAP_empty(OrFr_members(LOCAL_top_or_fr))) {
      FREE_OR_FRAME(LOCAL_top_or_fr);
    } else UNLOCK_OR_FRAME(LOCAL_top_or_fr);
    SCH_update_local_or_tops();
  }
  aux_frame = leftmost_cp->cp_or_fr;
  LOCK_OR_FRAME(aux_frame);
  if (LOCAL_cut_cp != NULL) {
    UNLOCK_OR_FRAME(aux_frame);
    CUT_free_pending_solutions(aux_solution);
    /*CUT_move_up_to_cut_cp();*/
    /*CUT_clear_cut_cp();*/
  } else {
    if (aux_solution != NULL) 
      CUT_put_inherted_solutions(aux_solution, aux_frame);
    UNLOCK_OR_FRAME(aux_frame);
  }
  B = LOCAL_top_cp; 
  return;

not_leftmost:
  /* performing a cut in the shared region, not being leftmost */
  BITMAP_clear(aux_bitmap);
  BITMAP_insert(aux_bitmap, worker_id);
  while (LOCAL_top_cp != leftmost_cp) {
    LOCK_OR_FRAME(LOCAL_top_or_fr);
    OrFr_alternative(LOCAL_top_or_fr) = NULL;
    BITMAP_difference(aux_bitmap, OrFr_members(LOCAL_top_or_fr), aux_bitmap);
    for (i = 0; i < number_workers; i++)
      if (BITMAP_member(aux_bitmap, i)) {
        CUT_make_cut_request(i, leftmost_cp->cp_b);
      } 
    aux_solution = CUT_put_not_cut_solutions_together(aux_solution);
    if (OrFr_pend_cut_cp(LOCAL_top_or_fr) != NULL) cut_more = 0;
    BITMAP_copy(aux_bitmap, OrFr_members(LOCAL_top_or_fr));
    BITMAP_delete(OrFr_members(LOCAL_top_or_fr), worker_id);
    if (BITMAP_empty(OrFr_members(LOCAL_top_or_fr))) {
      FREE_OR_FRAME(LOCAL_top_or_fr);
    } else UNLOCK_OR_FRAME(LOCAL_top_or_fr);
    SCH_update_local_or_tops();
  }
  LOCK_OR_FRAME(LOCAL_top_or_fr);
  if (LOCAL_cut_cp != NULL) {
    UNLOCK_OR_FRAME(LOCAL_top_or_fr);
    CUT_free_pending_solutions(aux_solution);
    /*CUT_move_up_to_cut_cp();*/
    /*CUT_clear_cut_cp();*/
  } else {
    leftmost_cp = leftmost_cp->cp_b;
    if (aux_solution != NULL) 
      CUT_put_inherted_solutions(aux_solution, LOCAL_top_or_fr);
    OrFr_alternative(LOCAL_top_or_fr) = NULL;
    BITMAP_difference(aux_bitmap, OrFr_members(LOCAL_top_or_fr), aux_bitmap);
    for (i = 0; i < number_workers; i++)
      if (BITMAP_member(aux_bitmap, i)) {
        if (GLOBAL_branch_ltt(i, OrFr_depth(LOCAL_top_or_fr)) <= 
            GLOBAL_branch_ltt(worker_id, OrFr_depth(LOCAL_top_or_fr))) {
          CUT_make_cut_request(i, leftmost_cp); 
        } else if (GLOBAL_branch_cut(i, OrFr_depth(LOCAL_top_or_fr)) == 1) cut_more = 0; 
      }
    if (OrFr_pend_cut_cp(LOCAL_top_or_fr) != NULL) cut_more = 0;
    OrFr_pend_cut_cp(LOCAL_top_or_fr) = cut_cp;
    OrFr_pend_cut_ltt(LOCAL_top_or_fr) = GLOBAL_branch_ltt(worker_id, OrFr_depth(LOCAL_top_or_fr));
    UNLOCK_OR_FRAME(LOCAL_top_or_fr);
    if (cut_more) {
      BITMAP_copy(aux_bitmap, OrFr_members(LOCAL_top_or_fr));
      while (leftmost_cp != cut_cp) {
        aux_frame = leftmost_cp->cp_or_fr;
        leftmost_cp = leftmost_cp->cp_b;
        LOCK_OR_FRAME(aux_frame);
        OrFr_alternative(aux_frame) = NULL;
        BITMAP_difference(aux_bitmap, OrFr_members(aux_frame), aux_bitmap);
        for (i = 0; i < number_workers; i++)
          if (BITMAP_member(aux_bitmap, i)) {
            if (GLOBAL_branch_ltt(i, OrFr_depth(aux_frame)) < 
                GLOBAL_branch_ltt(worker_id, OrFr_depth(aux_frame))) {
              CUT_make_cut_request(i, leftmost_cp); 
            } else if (GLOBAL_branch_cut(i, OrFr_depth(aux_frame)) == 1) {
              UNLOCK_OR_FRAME(aux_frame);
              goto end_cut_more;
            }
	  }
        BITMAP_copy(aux_bitmap, OrFr_members(aux_frame));
        UNLOCK_OR_FRAME(aux_frame);
      }
    }
end_cut_more:
  }
  B = LOCAL_top_cp; 
  return;
}
#endif /* YAPOR */
