/*************************************************************************
*									 *
*	 YAP Prolog 	%W% %G% 					 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		Regs.h							 *
* mods:									 *
* comments:	YAP abstract machine registers				 *
* version:      $Id: Regs.h,v 1.1 1997/06/02 16:32:00 vsc Exp $	 *
*************************************************************************/


/*********  abstract machine registers **********************************/

#include "amidefs.h"

#define MaxTemps	512

#ifdef i386
#define PUSH_REGS 1
#undef  PUSH_X
#endif

#ifdef sparc
#undef  PUSH_REGS
#undef  PUSH_X
#endif

#ifdef __alpha
#undef  PUSH_REGS
#undef  PUSH_X
#endif

#ifdef _POWER
#undef  PUSH_REGS
#undef  PUSH_X
#endif

#ifdef mips
#undef  PUSH_REGS
#undef  PUSH_X
#endif

EXTERN void restore_machine_regs(void);
EXTERN void save_machine_regs(void);
EXTERN void restore_H(void);
EXTERN void save_H(void);
EXTERN void restore_B(void);
EXTERN void save_B(void);


typedef struct
  {
    CELL CreepFlag_REG;		/* 13                                         */
    CELL *HB_REG;		/* 4 heap (global) stack top at latest c.p.   */
#if defined(SBA) && defined(YAPOR)
    choiceptr BB_REG;		/* 4 local stack top at latest c.p.   */
#endif /* SBA && YAPOR */
    tr_fr_ptr TR_REG;		/* 24 top of trail                            */
    CELL *H_REG;		/* 25 top of heap (global)  stack             */
    choiceptr B_REG;		/* 26 latest choice point                     */
#ifdef  DEPTH_LIMIT
    CELL DEPTH_REG;		/* 27                                         */
#endif  /* DEPTH_LIMIT */
    yamop *CP_REG;		/* 28 continuation program counter            */
    yamop *P_REG;		/* 7 prolog machine program counter           */
    CELL *YENV_REG;		/* 5 current environment (may differ from ENV)*/
    CELL *S_REG;		/* 6 structure pointer                        */
    CELL *ENV_REG;		/* 1 current environment                      */
    CELL *ASP_REG;		/* 8 top of local       stack                 */
    CELL *H0_REG;		/* 2 base of heap (global) stack              */
    CELL *LCL0_REG;		/* 3 local stack base                         */
    CELL *AuxSp_REG;		/* 9 Auxiliary stack pointer                  */
    ADDR AuxTop_REG;		/* 10 Auxiliary stack top                     */
    ADDR HeapPlus_REG;		/* 11 To avoid any chock with HeapTop         */
    tr_fr_ptr MyTR_REG;		/* 12                                         */
    CELL *CreepCode_REG;	/* 14                                         */
    CELL *UndefCode_REG;	/* 15                                         */
    CELL *SpyCode_REG;		/* 16                                         */
    CELL *LockCode_REG;		/* 17                                         */
    CELL FlipFlop_REG;		/* 18                                         */
    Term TermDot_REG;		/* 19                                         */
    Term TermNil_REG;		/* 20                                         */
#ifdef COROUTINING
    CELL *WakeUpCode_REG;          /* 21                                      */
#ifdef MULTI_ASSIGNMENT_VARIABLES
    Term WokenGoals_REG;	   /* 22                                      */
    Term MutableList_REG;          /* 23                                      */
#endif
#endif
#ifdef YAPOR
#ifdef SBA
    CELL *h_freeze_register;
    choiceptr b_freeze_register;
    tr_fr_ptr tr_freeze_register;
    choiceptr BSEG_REG;
    struct or_frame *frame_head_REG, *frame_tail_REG;
    char *binding_array_REG;
    int  sba_offset_REG;
    int  sba_end_REG;
    int  sba_size_REG;
#endif /* SBA */
#endif /* YAPOR */
#if PUSH_REGS
    /* On a X86 machine, the best solution is to keep the
       X registers on a global variable, whose address is known between
       runs, and to push the remaining registers to the stack.

       On a register based machine, one can have a dedicated register,
       always pointing to the XREG global variable. This spends an
       extra register, but makes it easier to access X[1].
     */

#if PUSH_X
    Term XTERMS[MaxTemps];	/* 29                                    */

#define XREGS	  REGS.XTERMS

#endif
  }
REGSTORE;


extern REGSTORE *regp;

#if !PUSH_X

	/* keep X as a global variable */

Term XREGS[MaxTemps];		/* 29                                     */
#endif

#define REGS (*regp)

#else /* PUSH_REGS */

    Term X[MaxTemps];		/* 29                                     */

#define XREGS	  REGS.X

  }
REGSTORE;

extern REGSTORE REGS;
#endif /* PUSH_REGS */

#define MinTrailGap (sizeof(CELL)*1024)
#define MinHeapGap  (sizeof(CELL)*4096)
#define MinStackGap (sizeof(CELL)*8*1024)
extern int stack_overflows;


#define ENV  REGS.ENV_REG	/* current environment                    */
#define ASP  REGS.ASP_REG	/* top of local   stack                   */
#define H0   REGS.H0_REG	/* base of heap (global) stack            */
#define LCL0 REGS.LCL0_REG	/* local stack base                       */
#ifdef  COROUTINING
#define WokenGoals  REGS.WokenGoals_REG
#define WakeUpCode  REGS.WakeUpCode_REG
#define MutableList REGS.MutableList_REG
#endif

#if defined(__GNUC__) && defined(sparc)

#define P    REGS.P_REG		/* prolog machine program counter         */
#define YENV REGS.YENV_REG	/* current environment (may differ from   ENV)*/
#define S    REGS.S_REG		/* structure pointer                      */

register CELL *H asm ("g6");
register tr_fr_ptr TR asm ("g7");
#ifdef __svr4__
register choiceptr B asm ("g5");
#else
#define  B         REGS.B_REG	/* latest choice point            */
#endif
#define CP         REGS.CP_REG	/* continuation   program counter         */
#define HB         REGS.HB_REG	/* heap (global) stack top at time of latest c.p. */
#define CreepFlag  REGS.CreepFlag_REG

#define BACKUP_MACHINE_REGS()           \
  CELL     *BK_H = H;                   \
  choiceptr BK_B = B;                   \
  tr_fr_ptr BK_TR = TR

#define RECOVER_MACHINE_REGS()          \
  H = BK_H;                             \
  B = BK_B;                             \
  TR = BK_TR

EXTERN inline void save_machine_regs(void) {
  REGS.H_REG   = H;
  REGS.TR_REG  = TR;
#ifdef __svr4__
  REGS.B_REG  = B;
#endif
}

EXTERN inline void restore_machine_regs(void) {
  H = REGS.H_REG;
  TR = REGS.TR_REG;
#ifdef __svr4__
  B = REGS.B_REG;
#endif
}


#define BACKUP_H()  CELL *BK_H = H

#define RECOVER_H()  H = BK_H

EXTERN inline void save_H(void) {
  REGS.H_REG   = H;
}

EXTERN inline void restore_H(void) {
  H = REGS.H_REG;
}

#define BACKUP_B()  choiceptr BK_B = B

#define RECOVER_B()  B = BK_B

EXTERN inline void save_B(void) {
#ifdef __svr4__
  REGS.B_REG   = B;
#endif
}

EXTERN inline void restore_B(void) {
#ifdef __svr4__
  B = REGS.B_REG;
#endif
}

#elif defined(__GNUC__) && defined(__alpha)

#define P               REGS.P_REG	/* prolog machine program counter */
#define YENV            REGS.YENV_REG	/* current environment (may differ from ENV) */
register CELL *H asm ("$9");
register CELL *HB asm ("$10");
register choiceptr B asm ("$11");
register yamop *CP asm ("$12");
register CELL *S asm ("$13");
register CELL CreepFlag asm ("$14");
register tr_fr_ptr TR asm ("$15");

/* Interface with foreign code, make sure the foreign code sees all the
   registers the way they used to be */
#define BACKUP_MACHINE_REGS()           \
  CELL     *BK_H = H;                   \
  CELL     *BK_HB = HB;                 \
  choiceptr BK_B = B;                   \
  CELL      BK_CreepFlag = CreepFlag;   \
  yamop    *BK_CP = CP;                 \
  tr_fr_ptr BK_TR = TR

#define RECOVER_MACHINE_REGS()          \
  H = BK_H;                             \
  HB = BK_HB;                           \
  B = BK_B;                             \
  CreepFlag = BK_CreepFlag;             \
  CP = BK_CP;                           \
  TR = BK_TR

EXTERN inline void save_machine_regs(void) {
  REGS.H_REG   = H;
  REGS.HB_REG = HB;
  REGS.B_REG   = B;
  REGS.CP_REG = CP;
  REGS.CreepFlag_REG = CreepFlag;
  REGS.TR_REG  = TR;
}

EXTERN inline void restore_machine_regs(void) {
  H = REGS.H_REG;
  HB = REGS.HB_REG;
  B = REGS.B_REG;
  CP = REGS.CP_REG;
  CreepFlag = REGS.CreepFlag_REG;
  TR = REGS.TR_REG;
}

#define BACKUP_H()  CELL *BK_H = H

#define RECOVER_H()  H = BK_H

EXTERN inline void save_H(void) {
  REGS.H_REG   = H;
}

EXTERN inline void restore_H(void) {
  H = REGS.H_REG;
}

#define BACKUP_B()  choiceptr BK_B = B

#define RECOVER_B()  B = BK_B

EXTERN inline void save_B(void) {
  REGS.B_REG   = B;
}

EXTERN inline void restore_B(void) {
  B = REGS.B_REG;
}

#elif defined(__GNUC__) && defined(mips)

#define P               REGS.P_REG	/* prolog machine program counter */
#define YENV            REGS.YENV_REG	/* current environment (may differ from ENV) */
register CELL *H asm ("$16");
register CELL *HB asm ("$17");
register choiceptr B asm ("$18");
register yamop *CP asm ("$19");
register CELL *S asm ("$20");
register CELL CreepFlag asm ("$21");
register tr_fr_ptr TR asm ("$22");

#define BACKUP_MACHINE_REGS()           \
  CELL     *BK_H = H;                   \
  CELL     *BK_HB = HB;                 \
  choiceptr BK_B = B;                   \
  CELL      BK_CreepFlag = CreepFlag;   \
  yamop    *BK_CP = CP;                 \
  tr_fr_ptr BK_TR = TR

#define RECOVER_MACHINE_REGS()          \
  H = BK_H;                             \
  HB = BK_HB;                           \
  B = BK_B;                             \
  CreepFlag = BK_CreepFlag;             \
  CP = BK_CP;                           \
  TR = BK_TR

EXTERN inline void save_machine_regs(void) {
  REGS.H_REG   = H;
  REGS.HB_REG = HB;
  REGS.B_REG   = B;
  REGS.CP_REG  = CP;
  REGS.CreepFlag_REG = CreepFlag;
  REGS.TR_REG  = TR;
}

EXTERN inline void restore_machine_regs(void) {
  H = REGS.H_REG;
  HB = REGS.HB_REG;
  B = REGS.B_REG;
  CP = REGS.CP_REG;
  CreepFlag = REGS.CreepFlag_REG;
  TR = REGS.TR_REG;
}

#define BACKUP_H()  CELL *BK_H = H

#define RECOVER_H()  H = BK_H

EXTERN inline void save_H(void) {
  REGS.H_REG   = H;
}

EXTERN inline void restore_H(void) {
  H = REGS.H_REG;
}

#define BACKUP_B()  choiceptr BK_B = B

#define RECOVER_B()  B = BK_B

EXTERN inline void save_B(void) {
  REGS.B_REG = B;
}

EXTERN inline void restore_B(void) {
  B = REGS.B_REG;
}

#elif defined(__GNUC__) && defined(_POWER)

/* 

   Because of a bug in GCC, we should always start from the first available
   general register. According to rs6000.h, this is r13 everywhere
   except for svr4 machines, where r13 is fixed.

   If we don't do so, GCC will generate move multiple instructions for
   entering complex functions. These instructions will save and
   restore the global registers :-(.

 */
#ifndef __svr4__
register CELL CreepFlag  asm ("r13");
#else
register CELL CreepFlag  asm ("r21");
#endif
register CELL *H asm ("r14");
register CELL *HB asm ("r15");
register choiceptr B asm ("r16");
register yamop *CP asm ("r17");
register CELL *S asm ("r18");
register CELL *YENV asm ("r19");
register tr_fr_ptr TR asm ("r20");
#define P    REGS.P_REG		/* prolog machine program counter */

#define BACKUP_MACHINE_REGS()           \
  CELL     *BK_H = H;                   \
  CELL     *BK_HB = HB;                 \
  choiceptr BK_B = B;                   \
  CELL      BK_CreepFlag = CreepFlag;   \
  yamop    *BK_CP = CP;                 \
  tr_fr_ptr BK_TR = TR

#define RECOVER_MACHINE_REGS()          \
  H = BK_H;                             \
  HB = BK_HB;                           \
  B = BK_B;                             \
  CreepFlag = BK_CreepFlag;             \
  CP = BK_CP;                           \
  TR = BK_TR

EXTERN inline void save_machine_regs(void) {
  REGS.CreepFlag_REG = CreepFlag;
  REGS.H_REG   = H;
  REGS.HB_REG = HB;
  REGS.B_REG   = B;
  REGS.CP_REG = CP;
  REGS.YENV_REG  = YENV;
  REGS.TR_REG  = TR;
}

EXTERN inline void restore_machine_regs(void) {
  CreepFlag = REGS.CreepFlag_REG;
  H = REGS.H_REG;
  HB = REGS.HB_REG;
  B = REGS.B_REG;
  CP = REGS.CP_REG;
  YENV = REGS.YENV_REG;
  TR = REGS.TR_REG;
}

#define BACKUP_H()  CELL *BK_H = H

#define RECOVER_H()  H = BK_H

EXTERN inline void save_H(void) {
  REGS.H_REG   = H;
}

EXTERN inline void restore_H(void) {
  H = REGS.H_REG;
}

#define BACKUP_B()  choiceptr BK_B = B

#define RECOVER_B()  B = BK_B

EXTERN inline void save_B(void) {
  REGS.B_REG   = B;
}

EXTERN inline void restore_B(void) {
  B = REGS.B_REG;
}

#else

#define CP         REGS.CP_REG	/* continuation   program counter         */
#define P          REGS.P_REG	/* prolog machine program counter */
#define YENV       REGS.YENV_REG /* current environment (may differ from ENV) */
#define S          REGS.S_REG	/* structure pointer                      */
#define	H          REGS.H_REG	/* top of heap (global)   stack           */
#define B          REGS.B_REG	/* latest choice point            */
#define TR         REGS.TR_REG	/* top of trail                           */
#define HB         REGS.HB_REG	/* heap (global) stack top at time of latest c.p. */
#define CreepFlag REGS.CreepFlag_REG

#define BACKUP_MACHINE_REGS()

#define RECOVER_MACHINE_REGS()

EXTERN inline void save_machine_regs(void) {
}

EXTERN inline void restore_machine_regs(void) {
}

#define BACKUP_H()

#define RECOVER_H()

EXTERN inline void save_H(void) {
}

EXTERN inline void restore_H(void) {
}

#define BACKUP_B()

#define RECOVER_B()

EXTERN inline void save_B(void) {
}

EXTERN inline void restore_B(void) {
}

#endif

#define	AuxSp     REGS.AuxSp_REG
#define	AuxTop    REGS.AuxTop_REG
#define	HeapPlus  REGS.HeapPlus_REG	/*To avoid any chock with HeapTop */
#define MyTR	  REGS.MyTR_REG
#define CreepCode REGS.CreepCode_REG
#define UndefCode REGS.UndefCode_REG
#define SpyCode   REGS.SpyCode_REG
#define LockCode  REGS.LockCode_REG
#define FlipFlop  REGS.FlipFlop_REG
#define TermDot	  REGS.TermDot_REG
#define TermNil	  REGS.TermNil_REG
#define DEPTH	  REGS.DEPTH_REG
#ifdef YAPOR
#ifdef SBA
#define H_FZ          REGS.h_freeze_register
#define B_FZ          REGS.b_freeze_register
#define TR_FZ         REGS.tr_freeze_register
#define BSEG	      REGS.BSEG_REG
#define binding_array REGS.binding_array_REG
#define sba_offset    REGS.sba_offset_REG
#define sba_end       REGS.sba_end_REG
#define sba_size      REGS.sba_size_REG
#define frame_head    REGS.frame_head_REG
#define frame_tail    REGS.frame_tail_REG
#endif
#endif 

#define REG_SIZE	sizeof(REGS)/sizeof(CELL *)

#define ARG1	XREGS[1]
#define ARG2	XREGS[2]
#define ARG3    XREGS[3]
#define ARG4	XREGS[4]
#define ARG5	XREGS[5]
#define ARG6	XREGS[6]
#define ARG7	XREGS[7]
#define ARG8	XREGS[8]
#define ARG9	XREGS[9]
#define ARG10	XREGS[10]
#define ARG11	XREGS[11]
#define ARG12	XREGS[12]
#define ARG13	XREGS[13]
#define ARG14	XREGS[14]
#define ARG15	XREGS[15]
#define ARG16	XREGS[16]

#define cut_succeed()	return( ( B = B->cp_b, 1 ))
#define cut_fail()	return( ( B = B->cp_b, 0 ))


#define EXTRA_CBACK_ARG(Arity,Offset)  (((CELL *)(B+1))[Arity+Offset-1])

/* by default, define HBREG to be HB */

#define HBREG HB

#ifdef YAPOR
#ifdef SBA
#define BB            REGS.BB_REG
#define BBREG         BB
#endif
#endif 

/* use global stack for intermediate data-structures */
#define AllocTmpSpaceFromGlobal()           ((char *)H)

/* optimise register allocation */
#define PRECOMPUTE_REGADDRESS 1

/* aggregate several abstract machine operations in a single */
#define AGGREGATE_OPS 1

/* make standard registers globally accessible so that they are there
   when we come from a longjmp */
#if PUSH_REGS
/* In this case we need to initialise the abstract registers */
REGSTORE standard_regs;
#endif /* PUSH_REGS */

