/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		absmi.c							 *
* Last rev:								 *
* mods:									 *
* comments:	Portable abstract machine interpreter includes           *
*									 *
*************************************************************************/

#ifdef SCCS
static char SccsId[] = "%W% %G%";

#endif /* SCCS */

/***************************************************************
* Macros for register manipulation                             *
***************************************************************/
/*
 * Machine and compiler dependent definitions
 */
#ifdef __GNUC__

#ifdef sparc
#define SHADOW_P       1
#define SHADOW_Y       1
#define SHADOW_S       1
#define SHADOW_REGS    1
#define SHADOW_CP      1
#define SHADOW_HB      1
/*#define SHADOW_CrFl    1 Breaks alarm/3 */
#define USE_PREFETCH   1
#endif

#ifdef __alpha
#define SHADOW_P       1
#define SHADOW_Y       1
#define SHADOW_REGS    1
#define USE_PREFETCH   1
#endif

#ifdef mips
#define SHADOW_P       1
#define SHADOW_Y       1
#define SHADOW_REGS    1
#define USE_PREFETCH   1
#endif

#ifdef _POWER
#define SHADOW_P       1
#define SHADOW_REGS    1
#define USE_PREFETCH   1
#endif

#ifdef i386
#define      Y_IN_MEM  1
#define      S_IN_MEM  1
#define     TR_IN_MEM  1
#define HAVE_FEW_REGS  1
#define LIMITED_PREFETCH   1
#ifdef BP_FREE
/***************************************************************
* Use bp as PREG for X86 machines		               *
***************************************************************/
register  void* P1REG asm ("bp"); /* can't use yamop before Yap.h */
#define PREG ((yamop *)P1REG)
#define NEEDS_TO_SET_PC 1
#endif /* BP_FREE */
#endif /* i386 */

#else /* other compilers */

#define S_IN_MEM       1

/* This works for xlc under AIX 3.2.5 */
#ifdef _IBMR2
#define SHADOW_P       1
#define SHADOW_REGS    1
#define SHADOW_S       1
#endif

#ifdef i386
#define Y_IN_MEM       1
#define S_IN_MEM       1
#define TR_IN_MEM      1
#define HAVE_FEW_REGS  1
#endif

#ifdef mips
#define SHADOW_P       1
#define SHADOW_Y       1
#define SHADOW_S       1
#define SHADOW_CP      1
#define SHADOW_HB      1
#define SHADOW_REGS    1
#define USE_PREFETCH   1
#endif

#endif /* __GNUC__ */

#ifdef DEBUG
/**********************************************************************
 *                                                                    *
 *                 Debugging Auxiliary variables                      *
 *                                                                    *
 **********************************************************************/
#include <stdio.h>

#endif
#include "Yap.h"
#include "clause.h"
#include "eval.h"
#ifdef YAPOR
#include "or.macros.h"
#endif	/* YAPOR */
#ifdef TABLING
#include "tab.macros.h"
#endif /* TABLING */
#ifdef LOW_LEVEL_TRACER
#include "tracer.h"
#endif

int  STD_PROTO(IUnify_complex, (CELL *, CELL *,CELL *));

#ifdef ANALYST

static char *op_names[_std_top + 1] =
{
#define OPCODE(OP,TYPE) #OP
#include "YapOpcodes.h"
#undef  OPCODE
};

#endif


#if PUSH_REGS

/***************************************************************
* Trick to copy REGS into absmi local environment              *
***************************************************************/

/* regp is a global variable */

inline EXTERN void
init_absmi_regs(REGSTORE * absmi_regs)
{
  memcpy(absmi_regs, regp, sizeof(REGSTORE));
}

inline EXTERN void
restore_absmi_regs(REGSTORE * old_regs)
{
  memcpy(old_regs, regp, sizeof(REGSTORE));
  regp = old_regs;
}
#endif /* PUSH_REGS */

/*****************************************************************

   Machine Dependent stuff 

******************************************************************/

#ifdef LONG_LIVED_REGISTERS

#define BEGP(TMP)

#define ENDP(TMP)

#define BEGD(TMP)

#define ENDD(TMP)

#else

#define BEGP(TMP) { register CELL *TMP

#define ENDP(TMP) }

#define BEGD(TMP) { register CELL TMP

#define ENDD(TMP) }

#endif /* LONG_LIVED_REGISTERS */

#define BEGCHO(TMP) { register choiceptr TMP

#define ENDCHO(TMP) }

/***************************************************************
* Y is usually, but not always, a register. This affects       *
* choicepoints                                                 *
***************************************************************/

#if Y_IN_MEM

#define CACHE_Y(A) { register CELL *S_Y = ((CELL *)(A))

#define ENDCACHE_Y() Y = S_Y; }

#define B_Y   ((choiceptr)(S_Y))

#else

#define S_Y   (Y)

#define B_Y ((choiceptr)(Y))

#define CACHE_Y(A) { Y = ((CELL *)(A))

#define ENDCACHE_Y() }

#endif

#if Y_IN_MEM

#define CACHE_Y_AS_ENV(A) { register CELL *E_Y = (A)

#define WRITEBACK_Y_AS_ENV()   Y = E_Y

#define ENDCACHE_Y_AS_ENV() }

#else

#define E_Y (Y)

#define WRITEBACK_Y_AS_ENV()   

#define CACHE_Y_AS_ENV(A) { Y = (A)

#define ENDCACHE_Y_AS_ENV() }

#endif

#if S_IN_MEM

#define CACHE_A1()

#define CACHED_A1()	ARG1

#else

#define CACHE_A1()	(SREG = (CELL *)ARG1)

#define CACHED_A1()	((CELL)SREG)

#endif /* S_IN_MEM */

/***************************************************************
* TR is usually, but not always, a register. This affects      *
* backtracking                                                 *
***************************************************************/

#if TR_IN_MEM

#define CACHE_TR(A) { register tr_fr_ptr S_TR = (A)

#define RESTORE_TR()    TR = S_TR

#define ENDCACHE_TR() }

#else

#define S_TR  TR

#define CACHE_TR(A) { TR = (A)

#define RESTORE_TR()

#define ENDCACHE_TR() }

#endif

/***************************************************************
* S is usually, but not always, a register (X86 machines).     *
* This affects unification instructions                        *
***************************************************************/

#if S_IN_MEM

/* jump through hoops because idiotic gcc will go and read S from
   memory every time it uses S :-( */

#define CACHE_S() { register CELL * S_SREG;

#define ENDCACHE_S() }

#define READ_IN_S() S_SREG = SREG

#else

/* do nothing if you are on a decent machine */

#define CACHE_S() {

#define ENDCACHE_S() }

#define READ_IN_S()

#define S_SREG  SREG

#endif

#define WRITEBACK_S(X) SREG = (X)

/*****************************************************************

   End of Machine Dependent stuff 

******************************************************************/

/*****************************************************************

   Prefetch is a technique to obtain the place to jump to before actually
   executing instructions. It can speed up some machines, by having the
   next opcode in place before it is actually required for jumping.

******************************************************************/

#if USE_THREADED_CODE

#define DO_PREFETCH(TYPE) to_go = (void *)(NEXTOP(PREG,TYPE)->opc)

#define DO_PREFETCH_W(TYPE) to_go = (void *)(NEXTOP(PREG,TYPE)->u.o.opcw)

#if LIMITED_PREFETCH

#define ALWAYS_START_PREFETCH(TYPE) \
 { register void *to_go; DO_PREFETCH(TYPE)

#define ALWAYS_LOOKAHEAD(WHAT) \
 { register void *to_go = (void *)(WHAT)

#define ALWAYS_START_PREFETCH_W(TYPE) \
 { register void *to_go; DO_PREFETCH_W(TYPE)

#else

#define ALWAYS_START_PREFETCH(TYPE) {

#define ALWAYS_START_PREFETCH_W(TYPE) {

#define ALWAYS_LOOKAHEAD(WHERE) {

#endif

#if USE_PREFETCH

#define START_PREFETCH(TYPE) ALWAYS_START_PREFETCH(TYPE)

#define START_PREFETCH_W(TYPE) ALWAYS_START_PREFETCH_W(TYPE)

#define INIT_PREFETCH() \
     { register void *to_go;

#define PREFETCH_OP(X) \
     to_go = (void *)((X)->opc);

#else

#define START_PREFETCH(TYPE) {

#define START_PREFETCH_W(TYPE) {

#define INIT_PREFETCH() {

#define PREFETCH_OP(X)

#endif	/* USE_PREFETCH */

#else /* USE_THREADED_CODE */

#define ALWAYS_START_PREFETCH(TYPE) {

#define ALWAYS_START_PREFETCH_W(TYPE) {

#define ALWAYS_LOOKAHEAD(WHERE) {

#define START_PREFETCH(TYPE) {

#define START_PREFETCH_W(TYPE) {

#define INIT_PREFETCH() {

#define PREFETCH_OP(X)

#endif /* USE_THREADED_CODE */

#define ALWAYS_END_PREFETCH() }

#define ALWAYS_END_PREFETCH_W() }

#define END_PREFETCH() }

#define END_PREFETCH_W() }

/*****************************************************************

  How to jump to the next abstract machine instruction

******************************************************************/

#if USE_THREADED_CODE

#define JMPNext()						\
	goto *(void *)(PREG->opc)

#define JMPNextW()						\
	goto *(void *)(PREG->u.o.opcw)

#if USE_THREADED_CODE && LIMITED_PREFETCH

#define ALWAYS_GONext() goto *to_go;

#define ALWAYS_GONextW() goto *to_go;

#else

#define ALWAYS_GONext() JMPNext()

#define ALWAYS_GONextW() JMPNextW()

#endif

#if USE_PREFETCH

#define GONext() ALWAYS_GONext() 

#define GONextW() ALWAYS_GONextW() 

#else

#define GONext() JMPNext()

#define GONextW() JMPNextW()

#endif /* USE_PREFETCH */

#define Op(Label,Type)	 Label: START_PREFETCH(Type)

#define OpW(Label,Type)	 Label: START_PREFETCH_W(Type)

#define BOp(Label,Type)	 Label:

#define PBOp(Label,Type) Label: INIT_PREFETCH()

#define OpRW(Label,Type) Label:

#else /* do not use threaded code */

#define JMPNext()	goto nextop

#define JMPNextW()	goto nextop_write

#define GONext()	JMPNext()

#define GONextW()	JMPNextW()

#define ALWAYS_GONext() GONext()

#define ALWAYS_GONextW() GONextW()

#define Op(Label,Type)	 case _##Label: START_PREFETCH(Type)

#define OpW(Label,Type)	 case _##Label: START_PREFETCH_W(Type)

#define BOp(Label,Type)	 case _##Label:

#define PBOp(Label,Type) case _##Label: INIT_PREFETCH()

#define OpRW(Label,Type) case _##Label:

#endif

#define ENDOp() END_PREFETCH()

#define ENDOpW() END_PREFETCH_W()

#define ENDOpRW()

#define ENDBOp()

#define ENDPBOp() END_PREFETCH()

/**********************************************************************
 *                                                                    *
 *                           PC manipulation                          *
 *                                                                    *
 **********************************************************************/

/*
 * How to set up and move a PC in a nice and disciplined way
 * 
 */

typedef CELL label;

/* move PC */

#define ADJ(P,x)    (P)+ADJUST(sizeof(x))

/*
 * Lookup PredEntry Structure
 * 
 */

#define pred_entry(X)		((PredEntry *)(Unsigned(X)-(CELL)(&(((PredEntry *)NULL)->StateOfPred))))
#define pred_entry_from_code(X)		((PredEntry *)(Unsigned(X)-(CELL)(&(((PredEntry *)NULL)->CodeOfPred))))
#define PredFromDefCode(X)	((PredEntry *)(Unsigned(X)-(CELL)(&(((PredEntry *)NULL)->OpcodeOfPred))))
#define Flags(X)		pred_entry(X)->StateOfPred
#define PredCode(X)		pred_entry(X)->CodeOfPred
#define PredOpCode(X)		pred_entry(X)->OpcodeOfPred
#define TruePredCode(X)		pred_entry(X)->TrueCodeOfPred
#define PredFunctor(X)		pred_entry(X)->FunctorOfPred
#define PredArity(X)		pred_entry(X)->ArityOfPE
#define Module(X)		pred_entry(X)->ModuleOfPred

#define FlagOff(Mask,w) !(Mask&w)
#define FlagOn(Mask,w) (Mask&w)
#define ResetFlag(Mask,w) w &= ~Mask
#define SetFlag(Mask,w) w |= Mask

/**********************************************************************
 *                                                                    *
 *                         X register access                          *
 *                                                                    *
 **********************************************************************/

#if PRECOMPUTE_REGADDRESS

#define XREG(I)		(*(CELL *)(I))

#else

#define XREG(I)		XREGS[I]

#endif /* ALIGN_LONGS */

	/* The Unification Stack is the Auxiliary stack */

#define SP0 ((CELL *)AuxTop)
#define SP  AuxSp

/**********************************************************************
 *                                                                    *
 *                         RWREG Manipulatio                          *
 *                                                                    *
 **********************************************************************/

#define READ_MODE     1
#define WRITE_MODE    0

/**********************************************************************
 *                                                                    *
 *Setting Temporary Copies of Often Used WAM registers for efficiency *
 *                                                                    *
 **********************************************************************/

#ifdef SHADOW_P
#define NEEDS_TO_SET_PC 1
#endif

/* 
 * First, the PC
 */
#ifdef NEEDS_TO_SET_PC
#define set_pc()	PREG = P
#define save_pc()	P = PREG
#else
#define set_pc()
#define save_pc()
#define PREG            (P)
#endif

/* 
 * Next, Y
 */
#if SHADOW_Y
#define set_y()		Y = YENV
#define save_y()	YENV = Y
#else
#define set_y()
#define save_y()
#define Y               YENV
#endif

/* 
 * Next, CP
 */
#if SHADOW_CP
#define set_cp()	CPREG = CP
#define save_cp()	CP = CPREG
#else
#define set_cp()
#define save_cp()
#define CPREG           CP
#endif

/* 
 * Next, CP
 */
	/* Breaks alarm/3 */
#if SHADOW_CrFl
#define set_cf()	CFREG = CreepFlag
#define save_cf()	CreepFlag = CFREG
#else
#define set_cf()
#define save_cf()
#define CFREG           CreepFlag
#endif

/* Say which registers must be saved at register entry and register
 * exit */
#define setregs()                     \
	set_cf();                     \
	set_hb();                     \
	set_cp();                     \
	set_pc();                     \
        set_y()

#define saveregs()                     \
	save_cf();                     \
	save_hb();                     \
	save_cp();                     \
	save_pc();                     \
        save_y()

/************************************************************

Macros to check the limits of stacks

*************************************************************/

#if HAVE_SIGSEGV
#define OS_HANDLES_TR_OVERFLOW 1
#endif

#if OS_HANDLES_TR_OVERFLOW

#define check_trail()

#else

#define check_trail() if (Unsigned(TrailTop) - Unsigned(TR) < MinTrailGap) \
			goto notrailleft

#endif

#if defined(SBA) && defined(YAPOR)
#define check_stack(Label, GLOB)                             \
 if ( (Int)(Unsigned(E_Y) - CFREG) < (Int)(GLOB) &&          \
       (choiceptr)E_Y < B_FZ && E_Y > H_FZ       &&          \
       (GLOB) > H_FZ && (GLOB) < (CELL *)B_FZ) goto Label
#else
#define check_stack(Label, GLOB)                             \
 if ( (Int)(Unsigned(E_Y) - CFREG) < (Int)(GLOB) ) goto Label
#endif /* SBA && YAPOR */

/***************************************************************
* Macros for choice point manipulation                         *
***************************************************************/

/***************************************************************
* Store variable number of arguments in a choice point         *
***************************************************************/
/***
   pt1 points at the new choice point,
   pt0 points at XREGS[i]
   d0 is a counter
   The macro just pushes the arguments one by one to the local stack.
***/
#define store_args(arity)                                         \
                 BEGP(pt0);                                       \
		 pt0 = XREGS+(arity);                             \
		 while ( pt0 > XREGS )                            \
                   { register CELL x = pt0[0];                    \
                     S_Y = S_Y-1;			          \
                     --pt0;                                       \
                     (S_Y)[0] = x;	                          \
		   }                                              \
                 ENDP(pt0)

#define store_at_least_one_arg(arity)                             \
                 BEGP(pt0);                                       \
		 pt0 = XREGS+(arity);                             \
                 do { register CELL x = pt0[0];                   \
                     S_Y = (S_Y)-1;			          \
                     --pt0;                                       \
                     (S_Y)[0] = x;	                          \
		   }                                              \
		 while ( pt0 > XREGS );                           \
                 ENDP(pt0)

/***************************************************************
* Do the bulk of work in creating a choice-point               *
* AP: alternative pointer                                      *
***************************************************************/
/*
 * The macro just sets pt1 to point to the base of the choicepoint
 * and then fills in all the necessary fields
 */
#ifdef DEPTH_LIMIT
#define store_yaam_reg_cpdepth(CPTR) (CPTR)->cp_depth = DEPTH
#else
#define store_yaam_reg_cpdepth(CPTR)
#endif

#define store_yaam_regs(AP,I) \
                 { register yamop *x1 = (yamop *)(AP);           \
                   register CELL *x2 = ENV;			 \
                   /* Jump to CP_BASE */                         \
                   S_Y = (CELL *)((choiceptr)((S_Y)-(I))-1);     \
                   /* Save Information */                        \
		   HBREG = H;                                    \
                   B_Y->cp_tr = TR;				 \
                   B_Y->cp_h  = H;				 \
                   B_Y->cp_b  = B;				 \
                   store_yaam_reg_cpdepth(B_Y);                  \
                   B_Y->cp_cp = CPREG;				 \
                   B_Y->cp_ap = x1;				 \
                   B_Y->cp_env= x2;				 \
                 }

#define store_yaam_regs_for_either(AP,d0) \
                 pt1 --; /* Jump to CP_BASE */		         \
                 /* Save Information */                          \
		 HBREG = H;                                      \
                 pt1->cp_tr = TR;	                         \
                 pt1->cp_h = H;		                         \
		 pt1->cp_b = B;		                         \
                 store_yaam_reg_cpdepth(pt1);                    \
                 pt1->cp_cp = d0;                                \
                 pt1->cp_ap = (yamop *)AP;                       \
		 pt1->cp_env = ENV;

/***************************************************************
* Place B as the new place to cut to                           *
***************************************************************/
#define	  set_cut(E,B) (E)[E_CB] = (CELL)(B)

/***************************************************************
* Restore WAM registers from a choice point                    *
***************************************************************/

#ifdef DEPTH_LIMIT
#define restore_yaam_reg_cpdepth(CPTR) DEPTH = (CPTR)->cp_depth
#else
#define restore_yaam_reg_cpdepth(CPTR)
#endif

#ifdef YAPOR
#define YAPOR_update_alternative(CUR_ALT, NEW_ALT)  \
	  if (SCH_shared_cp()) {                    \
	    SCH_new_alternative(CUR_ALT, NEW_ALT);  \
	  } else
#else
#define YAPOR_update_alternative(CUR_ALT, NEW_ALT)
#endif /* YAPOR */

#ifdef SBA
#ifdef YAPOR
#ifndef FROZEN_REGS
#define FROZEN_REGS 1
#endif
#endif
#endif

#ifdef TABLING
#ifndef FROZEN_REGS
#define FROZEN_REGS 1
#endif
#endif

#if defined(YAPOR) && defined(SBA)
#define SET_BB(V)    BBREG = (V)
#else
#define SET_BB(V)
#endif


#ifdef FROZEN_REGS
#ifdef SBA
#define PROTECT_FROZEN_H(CPTR)                                  \
       ((Unsigned((Int)((CPTR)->cp_h)-(Int)(H_FZ)) <            \
	 Unsigned((Int)(B_FZ)-(Int)(H_FZ))) ?                   \
	(CPTR)->cp_h : H_FZ)
#define PROTECT_FROZEN_B(CPTR)                                  \
       ((Unsigned((Int)(CPTR)-(Int)(H_FZ)) <                    \
	 Unsigned((Int)(B_FZ)-(Int)(H_FZ)))  ?                  \
	(CPTR) : B_FZ)
	 /*
#define PROTECT_FROZEN_H(CPTR) ((CPTR)->cp_h > H_FZ && (CPTR)->cp_h < (CELL *)B_FZ ? (CPTR)->cp_h : H_FZ )

#define PROTECT_FROZEN_B(CPTR)  ((CPTR) < B_FZ && (CPTR) > (choiceptr)H_FZ ? (CPTR) : B_FZ )
	 */
#else
#define PROTECT_FROZEN_H(CPTR)  MAX_REF((CPTR)->cp_h, H_FZ)
#define PROTECT_FROZEN_B(CPTR)  MIN_REF(CPTR, B_FZ)
#endif /* SBA */
#else
#define PROTECT_FROZEN_H(CPTR)  (CPTR)->cp_h
#endif /* TABLING */

#define restore_yaam_regs(AP)                                    \
                 { register CELL *x1 = B_Y->cp_env;	         \
                   register yamop *x2;				 \
                   H = HBREG = PROTECT_FROZEN_H(B_Y);            \
		   restore_yaam_reg_cpdepth(B_Y);	         \
                   CPREG  = B_Y->cp_cp;		                 \
		   /* AP may depend on H */			 \
		   x2 = (yamop *)AP;		                 \
                   ENV    = x1;                                  \
                   B_Y->cp_ap = x2;                              \
                 }

/***************************************************************
* Restore variable number of arguments from a choice point     *
***************************************************************/
#define restore_args(Nargs)                                        \
                 BEGD(d0);                                         \
                 d0 = Nargs;                                       \
                 BEGP(pt0);                                        \
                 BEGP(pt1);                                        \
                 pt1 = (CELL *)(B_Y+1)+d0;                         \
                 pt0 = XREGS+1+d0;                                 \
	         while (pt0 > XREGS +1 )                           \
                   { register CELL x = pt1[-1];                    \
                     --pt0;                                        \
                     --pt1;                                        \
                     *pt0   = x;                                   \
		   }                                               \
                 ENDP(pt1);                                        \
                 ENDP(pt0);                                        \
                 ENDD(d0)

#define restore_at_least_one_arg(Nargs)                            \
                 BEGD(d0);                                         \
                 d0 = Nargs;                                       \
                 BEGP(pt0);                                        \
                 BEGP(pt1);                                        \
                 pt1 = (CELL *)(B_Y+1)+d0;                         \
                 pt0 = XREGS+1+d0;                                 \
                 do { register CELL x = pt1[-1];                   \
                     --pt0;                                        \
                     --pt1;                                        \
                     *pt0   = x;                                   \
		   }                                               \
	         while (pt0 > XREGS +1 );                          \
                 ENDP(pt1);                                        \
                 ENDP(pt0);                                        \
                 ENDD(d0)

/***************************************************************
* Execute trust to release YAAM registers and pop choice point *
***************************************************************/
#ifdef DEPTH_LIMIT
#define pop_yaam_reg_cpdepth(CPTR) DEPTH = (CPTR)->cp_depth
#else
#define pop_yaam_reg_cpdepth(CPTR)
#endif

#ifdef TABLING
#define TABLING_close_alt(CPTR) (CPTR)->cp_ap = NULL
#else
#define TABLING_close_alt(CPTR)
#endif /* TABLING */

#define pop_yaam_regs()                                           \
                 { register CELL *ptr1;                           \
                   H = PROTECT_FROZEN_H(B_Y);                     \
		   B = B_Y->cp_b;	                          \
                   pop_yaam_reg_cpdepth(B_Y);	                  \
		   CPREG = B_Y->cp_cp;		                  \
		   ptr1 = B_Y->cp_env;				  \
                   TABLING_close_alt(B_Y);	                  \
                   HBREG = PROTECT_FROZEN_H(B);		          \
                   ENV = ptr1;                                    \
                 }

#define pop_args(NArgs)                                           \
                 BEGD(d0);                                        \
                 d0 = (NArgs);                                    \
                 BEGP(pt0);                                       \
                 BEGP(pt1);                                       \
                 S_Y = (CELL *)(B_Y+1);	                          \
                 pt0 = XREGS + 1 ;                                \
                 pt1 = S_Y ;                                      \
		 while (pt0 < XREGS+1+d0)                         \
                   { register CELL x = pt1[0];                    \
                     pt1++;                                       \
                     pt0++;                                       \
                     pt0[-1] = x;                                 \
		   }                                              \
                 S_Y = pt1;					  \
                 ENDP(pt1);                                       \
                 ENDP(pt0);                                       \
                 ENDD(d0);

#define pop_at_least_one_arg(NArgs)                               \
                 BEGD(d0);                                        \
                 d0 = (NArgs);                                    \
                 BEGP(pt0);                                       \
                 BEGP(pt1);                                       \
                 pt1 = (CELL *)(B_Y+1);	                          \
                 pt0 = XREGS + 1 ;                                \
                 do { register CELL x = pt1[0];                   \
                     pt1++;                                       \
                     pt0++;                                       \
                     pt0[-1] = x;                                 \
		   }                                              \
		 while (pt0 < XREGS+1+d0);                        \
                 S_Y = pt1;	                                  \
                 ENDP(pt1);                                       \
                 ENDP(pt0);                                       \
                 ENDD(d0);

/**********************************************************************
 *                                                                    *
 *                    failure and backtracking                        *
 *                                                                    *
 **********************************************************************/

/* Failure can be called from two routines.
 * 
 * If from within the emulator, we should jump to the label fail.
 * 
 * If from within the complex-term unification routine, we should jump
 * to the label "cufail".
 * 
 */

#define FAIL(FAIL)			                         \
		goto FAIL

#ifdef TERM_EXTENSIONS
/* If co-routining is on, and we cannot unify terms, before failing we
 * need to test whether failure happened because we unified to a
 * suspended variable. If so, we should see whether we can execute the
 * action. If we cannot, we should fail, else we should wake up the
 * goal, and continue execution at the next reasonable point.
 */

#define FAIL_D(Lab, Action, D0, D1, Next, FAIL)                             \
     {                                                                      \
        Functor fun;                                                        \
      Lab:                                                                  \
        if (!IsAttachedTerm(D0)) goto FAIL;                                 \
        fun = FunctorOfTerm(D0);                                            \
	if ((D0 = attas[ExtFromFunctor(fun)].bind_op(Action,D0,D1)))        \
          { Next; }                                                         \
        else                                                                \
          goto FAIL;                                                        \
     }

#define FAIL_D0_D1(Lab, Next, f0, f1, FAIL)                                 \
     {                                                                      \
      Lab:                                                                  \
        if (IsAttachFunc(f0)) {                                             \
	  if (attas[ExtFromFunctor(f0)].bind_op(SBIND,d0,d1))               \
            { Next; }                                                       \
          else                                                              \
            goto FAIL;                                                      \
        } else if (IsAttachFunc(f1)) {                                      \
	  if (attas[ExtFromFunctor(f1)].bind_op(SBIND,d1,d0))               \
            { Next; }                                                       \
          else                                                              \
            goto FAIL;                                                      \
        }                                                                   \
        goto FAIL;                                                          \
      }

#define FAIL_DNL(Action, D0, D1, Next, FAIL)                                \
     {                                                                      \
        Functor fun;                                                        \
        if (!IsAttachedTerm(D0)) goto FAIL;                                 \
        fun = FunctorOfTerm(D0);                                            \
	if ((D0 = attas[ExtFromFunctor(fun)].bind_op(Action,D0,D1)))        \
          { Next; }                                                         \
        else                                                                \
          goto FAIL;                                                        \
     }

#define FAIL_D0_D1NL(Next, f0, f1, FAIL)                                    \
     {                                                                      \
        if (IsAttachFunc(f0)) {                                             \
	  if (attas[ExtFromFunctor(f0)].bind_op(SBIND,d0,d1))               \
            { Next; }                                                       \
          else                                                              \
            goto FAIL;                                                      \
        } else if (IsAttachFunc(f1)) {                                      \
	  if (attas[ExtFromFunctor(f1)].bind_op(SBIND,d1,d0))               \
            { Next; }                                                       \
          else                                                              \
            goto FAIL;                                                      \
        }                                                                   \
        goto FAIL;                                                          \
      }

#define FAIL_J(Lab, FAIL)       goto Lab;

#else

#define FAIL_D(Lab,Action, D0, D1, Next, FAIL)

#define FAIL_DNL(Action, D0, D1, Next, FAIL)    goto FAIL

#define FAIL_D0_D1(Lab,Next, f0, f1, FAIL)

#define FAIL_D0_D1NL(Next, f0, f1, FAIL)        goto FAIL

#define FAIL_J(Lab,FAIL)			goto FAIL

#endif /* TERM_EXTENSIONS */

/**********************************************************************
 *                                                                    *
 *                      unification routines                          *
 *                                                                    *
 **********************************************************************/

#define UnifyCells(a, b)                                          \
     if((a) > (b)) {                                              \
	if ((a)<=H) { Bind_Global((a),(CELL)(b)); }               \
	else if ((b)<= H) { Bind_Local((a),(CELL)(b)); }          \
	else { Bind_Local((b),(CELL) (a)); }                      \
     } else if((a) < (b)){                                        \
	if((b) <= H) { Bind_Global((b),(CELL) (a)); }             \
	else if ((a) <= H) { Bind_Local((b),(CELL) (a)); }        \
	else { Bind_Local((a),(CELL) (b)); }                      \
     }

/* unify two complex terms.
 * 
 * I use two stacks: one keeps the visited terms, and the other keeps the
 * terms to visit.
 * 
 * The terms-to-visit stack is used to implement traditional
 * recursion. The visited-terms-stack is used to link structures already
 * visited and allows unification of infinite terms
 * 
 */

#ifdef RATIONAL_TREES

#define UNWIND_CUNIF()                                        \
         while (visited < AuxSp) {                            \
            pt1 = (CELL *)visited[0];                         \
            *pt1 = visited[1];                                \
            visited += 2;                                     \
         }

#else
#define UNWIND_CUNIF()
#endif

#define UnifyBound(d0,d1)                                              \
  if (d0 == d1) GONext();                                              \
  if (IsPairTerm(d0)) {                                                \
    register CELL *pt0, *pt1;                                          \
    if (!IsPairTerm(d1)) { FAIL_DNL(SBIND, d1, d0, GONext(), fail); }  \
    pt0 = RepPair(d0);                                                 \
    pt1 = RepPair(d1);                                                 \
    save_hb();							       \
    if (IUnify_complex(pt0-1,pt0+1,pt1-1)) { GONext(); }               \
    else { FAIL(fail); }                                               \
  } else if (IsApplTerm(d0)) {                                         \
    register CELL *pt0, *pt1;                                          \
    if (!IsApplTerm(d1)) { FAIL_DNL(SBIND, d1, d0, GONext(), fail); }  \
    pt0 = RepAppl(d0);                                                 \
    pt1 = RepAppl(d1);                                                 \
    if (*pt0 != *pt1)                                                  \
      {FAIL_D0_D1NL(GONext(),(Functor)(*pt0), (Functor)(*pt1), fail);} \
    else d0 = *pt0;                                                    \
    if (IsExtensionFunctor((Functor)d0)) {                             \
      switch(d0) {                                                     \
      case (Int)FunctorDouble:                                         \
	if (EQUAL_FLOATS(pt0, pt1))                                    \
	  { GONext(); }                                                \
	else                                                           \
	  { FAIL(fail); }                                              \
      case (Int)FunctorGeneralRef:                                     \
	if (pt0[0] == pt1[1])                                          \
	  { GONext(); }                                                \
	else                                                           \
	  { FAIL(fail); }                                              \
	/* DBRef could only succeed if both terms are the same so it   \
	 * does not need a special case */                             \
      case (Int)FunctorLongInt:                                        \
	if (EQUAL_LONG_INTS(pt0, pt1))                                 \
	  { GONext(); }                                                \
	else                                                           \
	  { FAIL(fail); }                                              \
      default:                                                         \
        if (IsAttachFunc((Functor)d0) &&                               \
	    attas[ExtFromFunctor((Functor)d0)].bind_op(SBIND,AbsAppl(pt0),d1))  \
         { GONext(); }                                                 \
	else                                                           \
         { FAIL(fail); }                                               \
      }                                                                \
    }                                                                  \
    d0 = ArityOfFunctor((Functor)d0);                                  \
    save_hb();							       \
    if (IUnify_complex(pt0, pt0+d0, pt1)) { GONext(); }                \
    else { FAIL(fail); }                                               \
  }                                                                    \
  else { FAIL_DNL(SBIND, d1, d0, GONext(), fail); }

#ifdef TABLING
STD_PROTO(static inline void load_answer_trie, (ans_node_ptr, choiceptr));

static inline 
void load_answer_trie(ans_node_ptr NODE, choiceptr CP_PTR) {
  int subs_arity;
  CELL *subs_ptr;
  subs_ptr = (CELL *)(CONS_CP(CP_PTR) + 1);
  subs_arity = *subs_ptr;
  if (subs_arity) {
    int i, n_vars = 0;
    CELL *stack_vars, *stack_terms, *stack_refs, *stack_refs_base, *stack_top;
    ans_node_ptr aux_parent_node;
    stack_vars = stack_terms = AuxSp - MAX_TABLE_VARS;
    stack_top = (CELL *)TrailTop;

    /* load the new answer from the answer trie to the stack_terms */
    aux_parent_node = UNTAG_ANSWER_LEAF_NODE(TrNode_parent(NODE));
    do {
      STACK_PUSH(TrNode_entry(NODE), stack_terms, stack_top);
      NODE = aux_parent_node;
      aux_parent_node = TrNode_parent(aux_parent_node);
    } while (aux_parent_node != NULL);

    /* bind the substitution variables with the answer loaded in stack_terms */
    stack_refs_base = stack_refs = stack_terms;
    for (i = subs_arity; i >= 1; i--) {
      Term t = STACK_POP(stack_terms);
      int tag = t & TabTagBits;
      switch (tag) {
        case TabVarTagBits:
        { int var_index = VarIndexOfTableTerm(t);
          if (var_index == n_vars) {
            stack_vars[n_vars++] = *stack_refs;
	  }
          Bind((CELL *)*(subs_ptr + i), stack_vars[var_index]);
        } break;
        case TabNumberTagBits:
        case TabAtomTagBits:
          Bind((CELL *)*(subs_ptr + i), t);
          break;
        case TabPairTagBits:
          /* build a pair term as in function MkPairTerm */
          Bind((CELL *)*(subs_ptr + i), AbsPair(H));
          H += 2;
          STACK_PUSH(H - 1, stack_refs, stack_top);
          STACK_PUSH(H - 2, stack_refs, stack_top);
          break;
        case TabApplTagBits:
        { /* build a pair term as in function MkApplTerm */
          Functor f = (Functor)NonTagPart(t);
          int j, f_arity = ArityOfFunctor(f);
          Bind((CELL *)*(subs_ptr + i), AbsAppl(H));
          *H++ = (CELL) f;
          H += f_arity;
          for (j = 1; j <= f_arity; j++)
            STACK_PUSH(H - j, stack_refs, stack_top);
        } break;
        default:
          abort_optyap("unknown type tag in macro load_answer_trie");
      }
    }
    while (STACK_NOT_EMPTY(stack_refs, stack_refs_base)) {
      Term t = STACK_POP(stack_terms);
      int tag = t & TabTagBits;
      switch (tag) {
        case TabVarTagBits:
	{ int var_index = VarIndexOfTableTerm(t);
          if (var_index == n_vars) {
            stack_vars[n_vars++] = *stack_refs;
	  }
          *((CELL *)STACK_POP(stack_refs)) = stack_vars[var_index];
        } break;
        case TabNumberTagBits:
        case TabAtomTagBits:
          *((CELL *)STACK_POP(stack_refs)) = t;
          break;
        case TabPairTagBits:
          /* build a pair term as in function MkPairTerm */
          *((CELL *)STACK_POP(stack_refs)) = AbsPair(H);
          H += 2;
          STACK_PUSH(H - 1, stack_refs, stack_top);
          STACK_PUSH(H - 2, stack_refs, stack_top);
          break;
        case TabApplTagBits:
        { /* build a pair term as in function MkApplTerm */
          Functor f = (Functor)NonTagPart(t);
          int j, f_arity = ArityOfFunctor(f);
          *((CELL *)STACK_POP(stack_refs)) = AbsAppl(H);
          *H++ = (CELL) f;
          H += f_arity;
          for (j = 1; j <= f_arity; j++)
            STACK_PUSH(H - j, stack_refs, stack_top);
        } break;
        default:
          abort_optyap("unknown type tag in macro load_answer_trie");
      }
    }
  }
}


#define load_answer_trie_old(NODE, CP_PTR)                                              \
        { int subs_arity;                                                               \
          CELL *subs_ptr;                                                               \
          subs_ptr = (CELL *)(CONS_CP(CP_PTR) + 1);                                     \
          subs_arity = *subs_ptr;                                                       \
          if (subs_arity) {                                                             \
            int i, n_vars = 0;                                                          \
            CELL *stack_vars, *stack_terms, *stack_refs, *stack_refs_base, *stack_top;  \
            ans_node_ptr aux_parent_node;                                               \
            stack_vars = stack_terms = AuxSp - MAX_TABLE_VARS;                          \
            stack_top = (CELL *)TrailTop;                                               \
                                                                                        \
            /* load the new answer from the answer trie to the stack_terms */           \
            aux_parent_node = UNTAG_ANSWER_LEAF_NODE(TrNode_parent(NODE));              \
            do {                                                                        \
              STACK_PUSH(TrNode_entry(NODE), stack_terms, stack_top);                   \
              NODE = aux_parent_node;                                                   \
              aux_parent_node = TrNode_parent(aux_parent_node);                         \
            } while (aux_parent_node != NULL);                                          \
                                                                                        \
            /* bind the substitution variables with the answer loaded in stack_terms */ \
            stack_refs_base = stack_refs = stack_terms;                                 \
            for (i = subs_arity; i >= 1; i--) {                                         \
              {                                                                         \
                Term t = STACK_POP(stack_terms);                                        \
                int tag = t & TabTagBits;                                               \
                switch (tag) {                                                          \
                  case TabVarTagBits:                                                   \
                  { int var_index = VarIndexOfTableTerm(t);                             \
                    if (var_index == n_vars) {                                          \
                      stack_vars[n_vars++] = *stack_refs;                               \
	            }                                                                   \
                    Bind((CELL *)*(subs_ptr + i), stack_vars[var_index]);               \
                  } break;                                                              \
                  case TabNumberTagBits:                                                \
                  case TabAtomTagBits:                                                  \
                    Bind((CELL *)*(subs_ptr + i), t);                                   \
                    break;                                                              \
                  case TabPairTagBits:                                                  \
                    /* build a pair term as in function MkPairTerm */                   \
                    Bind((CELL *)*(subs_ptr + i), AbsPair(H));                          \
                    H += 2;                                                             \
                    STACK_PUSH(H - 1, stack_refs, stack_top);                           \
                    STACK_PUSH(H - 2, stack_refs, stack_top);                           \
                    break;                                                              \
                  case TabApplTagBits:                                                  \
                  { /* build a pair term as in function MkApplTerm */                   \
                    Functor f = (Functor)NonTagPart(t);                                 \
                    int j, f_arity = ArityOfFunctor(f);                                 \
                    Bind((CELL *)*(subs_ptr + i), AbsAppl(H));                          \
                    *H++ = (CELL) f;                                                    \
                    H += f_arity;                                                       \
                    for (j = 1; j <= f_arity; j++)                                      \
                      STACK_PUSH(H - j, stack_refs, stack_top);                         \
                  } break;                                                              \
                  default:                                                              \
                    abort_optyap("unknown type tag in macro load_answer_trie");         \
                }                                                                       \
	      }                                                                         \
              while (STACK_NOT_EMPTY(stack_refs, stack_refs_base)) {                    \
                Term t = STACK_POP(stack_terms);                                        \
                int tag = t & TabTagBits;                                               \
                switch (tag) {                                                          \
                  case TabVarTagBits:                                                   \
	          { int var_index = VarIndexOfTableTerm(t);                             \
                    if (var_index == n_vars) {                                          \
                      stack_vars[n_vars++] = *stack_refs;                               \
	            }                                                                   \
                    *((CELL *)STACK_POP(stack_refs)) = stack_vars[var_index];           \
                  } break;                                                              \
                  case TabNumberTagBits:                                                \
                  case TabAtomTagBits:                                                  \
                    *((CELL *)STACK_POP(stack_refs)) = t;                               \
                    break;                                                              \
                  case TabPairTagBits:                                                  \
                    /* build a pair term as in function MkPairTerm */                   \
                    *((CELL *)STACK_POP(stack_refs)) = AbsPair(H);                      \
                    H += 2;                                                             \
                    STACK_PUSH(H - 1, stack_refs, stack_top);                           \
                    STACK_PUSH(H - 2, stack_refs, stack_top);                           \
                    break;                                                              \
                  case TabApplTagBits:                                                  \
                  { /* build a pair term as in function MkApplTerm */                   \
                    Functor f = (Functor)NonTagPart(t);                                 \
                    int j, f_arity = ArityOfFunctor(f);                                 \
                    *((CELL *)STACK_POP(stack_refs)) = AbsAppl(H);                      \
                    *H++ = (CELL) f;                                                    \
                    H += f_arity;                                                       \
                    for (j = 1; j <= f_arity; j++)                                      \
                      STACK_PUSH(H - j, stack_refs, stack_top);                         \
                  } break;                                                              \
                  default:                                                              \
                    abort_optyap("unknown type tag in macro load_answer_trie");         \
                }                                                                       \
              }                                                                         \
            }                                                                           \
          }                                                                             \
        }
#endif /* TABLING */

/* 
 * Next, HB
 */
#if SHADOW_HB
#undef HBREG
#define set_hb()	HBREG = HB
#define save_hb()	HB = HBREG
#else
#define set_hb()
#define save_hb()
#endif

