/*************************************************************************
*									 *
*	 YAP Prolog    @(#)amidefs.h	1.3 3/15/90
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		amidefs.h						 *
* Last rev:								 *
* mods:									 *
* comments:	Abstract machine peculiarities				 *
*									 *
*************************************************************************/

/*
   Possible arguments to YAP emulator:
   AREG describes an A or X register;
   YREG describes an Y register
   COUNT is a small number (eg, number of arguments to a choicepoint,
   number of permanent variables in a environment
*/

typedef OPREG  AREG;
typedef OPREG  YREG;
typedef OPREG  COUNT;


/*

   Types of possible YAAM instructions.

   The meaning of the symbols in a abstract machine instruction is:

   c: constant
   d: predicate definition
   f: functor
   n: small number
   l: label
   x: argument or temporary register
   y: environment slot

*/
typedef struct yami {
  OPCODE opc;
  union {
       struct {
	 CELL next;
       } e;
       struct {
	 CELL                c;
	 CELL next;
       } c;
       struct {
	 CELL                c;
	 CODEADDR            l1;
	 CODEADDR            l2;
	 CELL next;
       } cll;
       struct {
	 CODEADDR            d;
	 CELL next;
       } d;
       struct {
	 CELL                f[sizeof(Float)/sizeof(CELL)];
	 CELL next;
       } F;
       struct {
	 Int                 i;
	 CELL next;
       } I;
       struct {
	 Functor             f;
	 CELL next;
       } f;
       struct {
	 Functor             f;
	 CODEADDR            l1;
	 CODEADDR            l2;
	 CELL next;
       } fll;
       struct {
	 CODEADDR            l;
	 CELL next;
       } l;
       struct {
#ifdef YAPOR
         COUNT             ltt;  /* alternatives left to try */
	 COUNT             cut;  /* 1 or 0 (have cuts/no cuts) */
         COUNT             seq;  /* is seq or not */
#endif /* YAPOR */
#ifdef TABLING
         struct table_entry *te; /* pointer to table entry */
#endif /* TABLING */
	 COUNT               s;
	 CODEADDR	     p;
	 CODEADDR            d;
	 CELL next;
       } ld;
       struct {
#ifdef YAPOR
         COUNT             ltt;
	 COUNT             cut;
         COUNT             seq;  /* is seq or not */
#endif /* YAPOR */
#ifdef TABLING
         struct table_entry *te; /* pointer to table entry */
#endif /* TABLING */
	 COUNT               s;
	 CODEADDR	     p;
	 CODEADDR            d;
	 CODEADDR            bl;
	 CELL next;
       } ldl;
       struct {
	 CODEADDR            l;
	 SMALLUNSGN          flags;
	 CELL next;
       } lf;
       struct {
#ifdef YAPOR
         COUNT               ltt;
	 COUNT               cut;
         COUNT               seq;  /* is seq or not */
#endif /* YAPOR */
#ifdef TABLING
         struct table_entry *te; /* pointer to table entry */
#endif /* TABLING */
	 COUNT               s;
	 CODEADDR	     p;
	 CODEADDR            d;
	 COUNT               extra;
	 CELL next;
       } lds;
       struct {
	 CODEADDR            l1;
	 CODEADDR            l2;
	 CODEADDR            l3;
	 CELL next;
       } lll;
       struct {
#ifdef YAPOR
         COUNT               ltt;
	 COUNT               cut;
	 COUNT               seq;
#endif /* YAPOR */
#ifdef TABLING
         struct table_entry *te; /* pointer to table entry */
#endif /* TABLING */
	 COUNT               s;
	 CODEADDR	     p;
	 CODEADDR            l1;
	 CODEADDR            l2;
	 CODEADDR            l3;
	 CELL next;
       } slll;
       struct {
	 CODEADDR            l1;
	 CODEADDR            l2;
	 CODEADDR            l3;
	 CODEADDR            l4;
	 CELL next;
       } llll;
       struct {
	 CODEADDR            l;
	 AREG                x1;
	 AREG                x2;
	 AREG                flags;
	 CELL next;
       } lxx;
       struct {
	 CODEADDR            l;
	 AREG                x;
	 YREG                y;
	 AREG                flags;
	 CELL next;
       } lxy;
       struct {
	 CODEADDR            l;
	 AREG                y1;
	 YREG                y2;
	 AREG                flags;
	 CELL next;
       } lyy;
       struct {
	 OPCODE              pop;
	 CODEADDR            l1;
	 CODEADDR            l2;
	 CODEADDR            l3;
	 CODEADDR            l4;
	 CELL next;
       } ollll;
       struct {
	 OPCODE              opcw;
	 CELL next;
       } o;
       struct {
	 OPCODE              opcw;
	 CELL                c;
	 CELL next;
       } oc;
       struct {
	 OPCODE              opfw;
	 CELL                f[sizeof(Float)/sizeof(CELL)];
	 CELL next;
       } oF;
       struct {
	 OPCODE              opIw;
	 Int                 i;
	 CELL next;
       } oI;
       struct {
	 OPCODE              opcw;
	 Functor             f;
	 CELL next;
       } of;
       struct {
	 OPCODE              opcw;
	 COUNT               s;
	 CELL                c;
	 CELL next;
       } osc;
       struct {
	 OPCODE              opcw;
	 COUNT               s;
	 CELL next;
       } os;
       struct {
	 OPCODE              opcw;
	 AREG                x;
	 CELL next;
       } ox;
       struct {
	 OPCODE              opcw;
	 AREG                xl;
	 AREG                xr;
	 CELL next;
       } oxx;
       struct {
	 OPCODE              opcw;
	 YREG                y;
	 CELL next;
       } oy;
       struct {
	 COUNT               s;
	 CELL next;
       } s;
       struct {
	 COUNT               s;
	 CELL                c;
	 CELL next;
       } sc;
       struct {
	 COUNT               s;
	 CODEADDR            d;
	 CODEADDR            l;
	 CELL next;
       } sdl;
       struct {
	 COUNT               s;
	 CODEADDR            l;
	 CELL next;
       } sl;
       struct {
#ifdef YAPOR
         COUNT             ltt;
	 COUNT             cut;
         COUNT             seq;
#endif /* YAPOR */
	 COUNT               s;
	 CODEADDR            l;
	 CODEADDR            l2;
	 CELL next;
       } sla; /* also check env for yes code before making any changes */
       struct {
	 AREG                x;
	 CELL next;
       } x;
       struct {
	 AREG                x;
	 CELL                c;
	 CELL next;
       } xc;
       struct {
	 AREG                x;
	 CELL                f[sizeof(Float)/sizeof(CELL)];
	 CELL next;
       } xF;
       struct {
	 AREG                x;
	 Int                 i;
	 CELL next;
       } xI;
       struct {
	 AREG                x;
	 Functor             f;
	 CELL next;
       } xf;
       struct {
	 AREG                xl;
	 AREG                xr;
	 CELL next;
       } xx;
       struct {
	 AREG                x;
	 AREG                x1;
	 AREG                x2;
	 CELL next;
       } xxx;
       struct {
	 AREG                x;
	 Int                 c;
	 AREG                xi;
	 CELL next;
       } xcx, xxc;
       struct {
	 AREG                x;
	 YREG                y;
	 CELL next;
       } xy;
       struct {
	 YREG                y;
	 CELL next;
       } y;
       struct {
	 YREG                y;
	 AREG                x;
	 CELL next;
       } yx;
       struct {
	 YREG                y;
	 AREG                x1;
	 AREG                x2;
	 CELL next;
       } yxx;
       struct {
	 YREG                y;
	 Int                 c;
	 AREG                xi;
	 CELL next;
       } ycx, yxc;
     } u;
} yamop;

typedef yamop yamopp;

#define OPCR                opc
#define OPCW                u.ox.opcw


#define NEXTOP(V,TYPE)    ((yamop *)(&((V)->u.TYPE.next)))

/*
   This is a table with the codes for YAP instructions
*/
typedef enum {
#define OPCODE(OP,TYPE) _##OP
#include "YapOpcodes.h"
#undef  OPCODE
} op_numbers;


#define _std_top	_p_slr_y_cv

typedef enum {
	_atom,
	_atomic,
	_integer,
	_compound,
	_float,
	_nonvar,
	_number,
	_var,
	_cut_by,
	_db_ref,
	_primitive,
	_equal,
	_dif,
	_eq,
	_arg,
	_functor,
	_plus,
	_minus,
	_times,
	_div,
	_and,
	_or,
	_sll,
	_slr
} basic_preds;

#if defined(TABLING) || defined(SBA)
typedef struct trail_frame {
  Term term;
  CELL value;
} *tr_fr_ptr;

#define TrailTerm(X)   ((X)->term)
#define TrailVal(X)    ((X)->value)
#else
typedef Term *tr_fr_ptr;

#define TrailTerm(X)   (*(X))
#define TrailOldVal(X) OOOOOOPS: this program should not compile
#endif /* TABLING || SBA  */


/*
   Choice Point Structure

   6 fixed fields (TR,AP,H,B,ENV,CP) plus arguments

*/
typedef struct choicept {
  tr_fr_ptr cp_tr;
  CELL *cp_h;
  struct choicept *cp_b;
#ifdef DEPTH_LIMIT
  CELL cp_depth;
#endif /* DEPTH_LIMIT */
  yamop *cp_cp;
#ifdef YAPOR
  int cp_lub;           /* local untried branches */
  struct or_frame *cp_or_fr;  /* or-frame pointer */
#endif /* YAPOR */
  yamop *cp_ap;
#ifdef __GNUC__
  CELL *cp_env;
  /* GNUCC understands empty arrays */
  CELL cp_args[0];
#define cp_a1		cp_args[0]
#define cp_a2		cp_args[1]
#define cp_a3		cp_args[2]
#define cp_a4		cp_args[3]
#define cp_a5		cp_args[4]
#define cp_a6		cp_args[5]
#define cp_a7		cp_args[6]
#define cp_a8		cp_args[7]
#else
  /* Otherwise, we need a very dirty trick to access the arguments */
  union {
    CELL *cp_uenv;
    CELL  cp_args[1];
  } cp_last;
#define cp_env		cp_last.cp_uenv
#define cp_a1		cp_last.cp_args[1]
#define cp_a2		cp_last.cp_args[2]
#define cp_a3		cp_last.cp_args[3]
#define cp_a4		cp_last.cp_args[4]
#define cp_a5		cp_last.cp_args[5]
#define cp_a6		cp_last.cp_args[6]
#define cp_a7		cp_last.cp_args[7]
#define cp_a8		cp_last.cp_args[8]
#endif
} *choiceptr;

/*
   Environment Structure (CP, E, and CUT_B). Yap always saves the B
   where to cut to, even if not needed.
*/
#define E_CP		-1
#define E_E		-2
#define E_CB		-3
#ifdef TABLING
#define E_B		-4
#ifdef  DEPTH_LIMIT
#define E_DEPTH         -5
#define EnvSizeInCells   5
#else
#define EnvSizeInCells   4
#endif  /* DEPTH_LIMIT */
#else   /* TABLING */
#ifdef  DEPTH_LIMIT
#define E_DEPTH         -4
#define EnvSizeInCells   4
#else
#define EnvSizeInCells   3
#endif  /* DEPTH_LIMIT */
#endif  /* TABLING */

#if MSHIFTOFFS
#define FixedEnvSize		EnvSizeInCells
#else
#define FixedEnvSize		(EnvSizeInCells*sizeof(CELL))
#endif
#define RealEnvSize	(EnvSizeInCells*sizeof(CELL))

#define ENV_Size(cp)     (((yamop *)((CODEADDR)(cp) - (CELL)NEXTOP((yamop *)NIL,sla)))->u.sla.s)
#define EnvSize(cp)     ((-ENV_Size(cp))/(OPREG)sizeof(CELL))
#define EnvPrdBMP(p)    ((CELL *)(((yamop *)((CODEADDR)(p) - (CELL)NEXTOP((yamop *)NIL,sla)))->u.sla.l2))
#define EnvBMapOffset(p)  ((p)+1)
#define EnvBMap(p)     EnvBMapOffset(EnvPrdBMP(p))
#define EnvPreg(p)     ((PredEntry *)(EnvPrdBMP(p) == NIL ? NIL : EnvPrdBMP(p)[0] ))
#define PredFromOr(p)     ((PredEntry *)(((CELL *)(p))[0] == NIL ? NIL : ((CELL *)(p))[0] ))

/* access to instructions */

#if USE_THREADED_CODE
extern void **ABSMI_OPCODES;

#define absmadr(i) ((OPCODE)(ABSMI_OPCODES[(i)]))
#else
#define absmadr(i) ((OPCODE)(i))
#endif


/* used to find out how many instructions of each kind are executed */
#ifdef ANALYST
extern int opcount[_std_top+1];
#endif /* ANALYST */
