/*************************************************************************
*									 *
*	 YAP Prolog 	@(#)c_interface.h	2.2			 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		c_interface.h						 *
* Last rev:	19/2/88							 *
* mods:									 *
* comments:	c_interface header file for YAP				 *
*									 *
*************************************************************************/

/*******************  IMPORTANT ********************
   Due to a limitation of the DecStation loader any function (including
   library functions) which is linked to yapn can not be called directly
   from C code loaded dynamically.
      To go around this problem we adopted the solution of calling such
   functions indirectly
****************************************************/


#if	defined(__STDC__)

#define PROTO(X,ARGS) X##ARGS

#else

#define PROTO(X,ARGS) X()

#endif

/* Type definitions */

typedef unsigned long CELL;  /* this is common to all current machines */

typedef int Bool;

typedef long int Int;

typedef double flt;

typedef CELL Term, Functor, Atom;

#define TRUE	1
#define FALSE	0
#define NIL     0

/* Primitive Functions */

/*  Term Deref(Term)  */
extern Term PROTO(C_A,(int));
#ifdef IndirectCalls
static Term (*CI_A)() = C_A;
#define A(I) (*CI_A)(I)
#else
#define A(I) C_A(I)
#endif
#define ARG1	A(1)
#define ARG2	A(2)
#define ARG3	A(3)
#define ARG4	A(4)
#define ARG5	A(5)
#define ARG6	A(6)
#define ARG7	A(7)
#define ARG8	A(8)
#define ARG9	A(9)
#define ARG10	A(10)
#define ARG11	A(11)
#define ARG12	A(12)
#define ARG13	A(13)
#define ARG14	A(14)
#define ARG15	A(15)
#define ARG16	A(16)

/*  Term Deref(Term)  */
extern Term PROTO(Deref,(Term));
#ifdef IndirectCalls
static Term (*CI_Deref)() = Deref;
#define Deref(T) (*CI_Deref)(T)
#endif

/*  Bool IsVarTerm(Term) */
extern Bool PROTO(C_IsVarTerm,(Term));
#ifdef IndirectCalls
static Bool (*CI_IsVarTerm)() = C_IsVarTerm;
#define IsVarTerm(T) (*CI_IsVarTerm)(T)
#else
#define IsVarTerm(T) C_IsVarTerm(T)
#endif

/*  Bool IsNonVarTerm(Term) */
extern Bool PROTO(C_IsNonVarTerm,(Term));
#ifdef IndirectCalls
static Bool (*CI_IsNonVarTerm)() = C_IsNonVarTerm;
#define IsNonVarTerm(T) (*CI_IsNonVarTerm)(T)
#else
#define IsNonVarTerm(T) C_IsNonVarTerm(T)
#endif

/*  Term  MkVarTerm()  */
extern Term PROTO(MkVarTerm,(void));
#ifdef IndirectCalls
static Term (*CI_MkVarTerm)() = C_MkVarTerm;
#define MVarTerm() (*CI_MkVarTerm)()
#else
#define MkVarTerm() C_MkVarTerm()
#endif

/*  Bool IsIntTerm(Term)  */
extern Bool PROTO(C_IsIntTerm,(Term));
#ifdef IndirectCalls
static Bool (*CI_IsIntTerm)() = C_IsIntTerm;
#define IsIntTerm(T) (*CI_IsIntTerm)(T)
#else
#define IsIntTerm(T) C_IsIntTerm(T)
#endif

/*  Bool IsFloatTerm(Term)  */
extern Bool PROTO(C_IsFloatTerm,(Term));
#ifdef IndirectCalls
static Bool (*CI_IsFloatTerm)() = C_IsFloatTerm;
#define IsFloatTerm(T) (*CI_IsFloatTerm)(T)
#else
#define IsFloatTerm(T) C_IsFloatTerm(T)
#endif

/*  Bool IsDbRefTerm(Term)  */
extern Bool PROTO(C_IsDbRefTerm,(Term));
#ifdef IndirectCalls
static Bool (*CI_IsDbRefTerm)() = C_IsDbRefTerm;
#define IsDbRefTerm(T) (*CI_IsDbRefTerm)(T)
#else
#define IsDbRefTerm(T) C_IsDbRefTerm(T)
#endif

/*  Bool IsAtomTerm(Term)  */
extern Bool PROTO(C_IsAtomTerm,(Term));
#ifdef IndirectCalls
static Bool (*CI_IsAtomTerm)() = C_IsAtomTerm;
#define IsAtomTerm(T) (*CI_IsAtomTerm)(T)
#else
#define IsAtomTerm(T) C_IsAtomTerm(T)
#endif

/*  Bool IsPairTerm(Term)  */
extern Bool PROTO(C_IsPairTerm,(Term));
#ifdef IndirectCalls
static Bool (*CI_IsPairTerm)() = C_IsPairTerm;
#define IsPairTerm(T) (*CI_IsPairTerm)(T)
#else
#define IsPairTerm(T) C_IsPairTerm(T)
#endif

/*  Bool IsApplTerm(Term)  */
extern Bool PROTO(C_IsApplTerm,(Term));
#ifdef IndirectCalls
static Bool (*CI_IsApplTerm)() = C_IsApplTerm;
#define IsApplTerm(T) (*CI_IsApplTerm)(T)
#else
#define IsApplTerm(T) C_IsApplTerm(T)
#endif

/*    Term MkIntTerm(Int)  */
extern Term PROTO(C_MkIntTerm,(Int));
#ifdef IndirectCalls
static Term (*CI_MkIntTerm)() = C_MkIntTerm;
#define MkIntTerm(T) (*CI_MkIntTerm)(T)
#else
#define MkIntTerm(T) C_MkIntTerm(T)
#endif

/*    Int  IntOfTerm(Term) */
extern Int PROTO(C_IntOfTerm,(Term));
#ifdef IndirectCalls
static Int (*CI_IntOfTerm)() = C_IntOfTerm;
#define IntOfTerm(T) (*CI_IntOfTerm)(T)
#else
#define IntOfTerm(T) C_IntOfTerm(T)
#endif

/*    Term MkFloatTerm(flt)  */
extern Term PROTO(C_MkFloatTerm,(flt));
#ifdef IndirectCalls
static Term (*CI_MkFloatTerm)() = C_MkFloatTerm;
#define MkFloatTerm(T) (*CI_MkFloatTerm)(T)
#else
#define MkFloatTerm(T) C_MkFloatTerm(T)
#endif

/*    flt  FloatOfTerm(Term) */
extern flt PROTO(C_FloatOfTerm,(Term));
#ifdef IndirectCalls
static flt (*CI_FloatOfTerm)() = C_FloatOfTerm;
#define FloatOfTerm(T) (*CI_FloatOfTerm)(T)
#else
#define FloatOfTerm(T) C_FloatOfTerm(T)
#endif

/*    Term MkAtomTerm(Atom)  */
extern Term PROTO(C_MkAtomTerm,(Atom));
#ifdef IndirectCalls
static Term (*CI_MkAtomTerm)() = C_MkAtomTerm;
#define MkAtomTerm(T) (*CI_MkAtomTerm)(T)
#else
#define MkAtomTerm(T) C_MkAtomTerm(T)
#endif

/*    Atom  AtomOfTerm(Term) */
extern Atom PROTO(C_AtomOfTerm,(Term));
#ifdef IndirectCalls
static Atom (*CI_AtomOfTerm)() = C_AtomOfTerm;
#define AtomOfTerm(T) (*CI_AtomOfTerm)(T)
#else
#define AtomOfTerm(T) C_AtomOfTerm(T)
#endif

/*    Atom  LookupAtom(char *) */
extern Atom PROTO(LookupAtom,(char *));
#ifdef IndirectCalls
static Atom (*CI_LookupAtom)() = LookupAtom;
#define LookupAtom(T) (*CI_LookupAtom)(T)
#endif

/*    char* AtomName(Atom) */
extern char *PROTO(C_AtomName,(Atom));
#ifdef IndirectCalls
static char *((*CI_AtomName)()) = C_AtomName;
#define AtomName(T) (*CI_AtomName)(T)
#else
#define AtomName(T) C_AtomName(T)
#endif

/*    Term  MkPairTerm(Term Head, Term Tail) */
extern Term PROTO(MkPairTerm,(Term,Term));
#ifdef IndirectCalls
static Term (*CI_MkPairTerm)() = C_MkPairTerm;
#define MkPairTerm(T1,T2) (*CI_MkPairTerm)(T1,T2)
#else
#define MkPairTerm(T1,T2) C_MkPairTerm(T1,T2)
#endif

/*    Term  HeadOfTerm(Term)  */
extern Term PROTO(C_HeadOfTerm,(Term));
#ifdef IndirectCalls
static Term (*CI_HeadOfTerm)() = C_HeadOfTerm;
#define HeadOfTerm(T) (*CI_HeadOfTerm)(T)
#else
#define HeadOfTerm(T) C_HeadOfTerm(T)
#endif

/*    Term  TailOfTerm(Term)  */
extern Term PROTO(C_TailOfTerm,(Term));
#ifdef IndirectCalls
static Term (*CI_TailOfTerm)() = C_TailOfTerm;
#define TailOfTerm(T) (*CI_TailOfTerm)(T)
#else
#define TailOfTerm(T) C_TailOfTerm(T)
#endif


/*    Term     MkApplTerm(Functor f, int n, Term[] args) */
extern Term PROTO(C_MkApplTerm,(Functor,int,Term *));
#ifdef IndirectCalls
static Term (*CI_MkApplTerm)() = C_MkApplTerm;
#define MkApplTerm(F,N,As) (*CI_MkApplTerm)(F,N,As)
#else
#define MkApplTerm(F,N,As) C_MkApplTerm(F,N,As)
#endif


/*    Functor  FunctorOfTerm(Term)  */
extern Functor PROTO(C_FunctorOfTerm,(Term));
#ifdef IndirectCalls
static Functor (*CI_FunctorOfTerm)() = C_FunctorOfTerm;
#define FunctorOfTerm(T) (*CI_FunctorOfTerm)(T)
#else
#define FunctorOfTerm(T) C_FunctorOfTerm(T)
#endif

/*    Term     ArgOfTerm(int argno,Term t) */
extern Term PROTO(C_ArgOfTerm,(int,Term));
#ifdef IndirectCalls
static Term (*CI_ArgOfTerm)() = C_ArgOfTerm;
#define ArgOfTerm(N,T) (*CI_ArgOfTerm)(N,T)
#else
#define ArgOfTerm(N,T) C_ArgOfTerm(N,T)
#endif

/*    Functor  MkFunctor(Atom a,int arity) */
extern Functor PROTO(C_MkFunctor,(Atom,int));
#ifdef IndirectCalls
static Functor (*CI_MkFunctor)() = C_MkFunctor;
#define MkFunctor(A,N) (*CI_MkFunctor)(A,N)
#else
#define MkFunctor(A,N) C_MkFunctor(A,N)
#endif

/*    Atom     NameOfFunctor(Functor) */
extern Atom PROTO(C_NameOfFunctor,(Functor));
#ifdef IndirectCalls
static Atom (*CI_NameOfFunctor)() = C_NameOfFunctor;
#define NameOfFunctor(T) (*CI_NameOfFunctor)(T)
#else
#define NameOfFunctor(T) C_NameOfFunctor(T)
#endif

/*    Int     ArityOfFunctor(Functor) */
extern Int PROTO(C_ArityOfFunctor,(Functor));
#ifdef IndirectCalls
static Int (*CI_ArityOfFunctor)() = C_ArityOfFunctor;
#define ArityOfFunctor(T) (*CI_ArityOfFunctor)(T)
#else
#define ArityOfFunctor(T) C_ArityOfFunctor(T)
#endif

/*  void ExtraSpace(void) */
extern void *PROTO(C_ExtraSpace,(void));
#ifdef IndirectCalls
static void *(*CI_ExtraSpace)() = C_ExtraSpace;
#define ExtraSpace() (*C_ExtraSpace)()
#else
#define ExtraSpace() C_ExtraSpace()
#endif

#define  PRESERVE_DATA(ptr, type) (ptr = (type *)ExtraSpace())
#define PRESERVED_DATA(ptr, type) (ptr = (type *)ExtraSpace())

/*   Int      unify(Term* a, Term *b) */
extern Int PROTO(C_unify,(Term *, Term *));
#ifdef IndirectCalls
static Int (*CI_unify)() = C_unify;
#define unify(T1,T2) (*CI_unify)(T1,T2)
#else
#define unify(T1,T2) C_unify(T1,T2)
#endif

/*  void UserCPredicate(char *name, int *fn(), int arity) */
extern void PROTO(UserCPredicate,(char *, int (*)(void), int));
#ifdef IndirectCalls
static void (*CI_UserCPredicate)() = UserCPredicate;
#define UserCPredicate(N,F,A) (*CI_UserCPredicate)(N,F,A)
#endif

/*  void UserBackCPredicate(char *name, int *init(), int *cont(), int
    arity, int extra) */
extern void PROTO(UserBackCPredicate,(char *, int (*)(void), int (*)(void), int, int));
#ifdef IndirectCalls
static void (*CI_UserBackCPredicate)() = UserBackCPredicate;
#define UserBackCPredicate(N,F,G,A,B) (*CI_UserBackCPredicate)(N,F,G,A,B)
#endif

/*  void CallProlog(Term t) */
extern void PROTO(C_CallProlog,(Term t));
#ifdef IndirectCalls
static void (*CI_CallProlog)() = C_CallProlog;
#define CallProlog(t) (*CI_CallProlog)(t)
#else
#define CallProlog(t) C_CallProlog(t)
#endif

/*  void cut_fail(void) */
extern Int PROTO(C_cut_fail,(void));
#ifdef IndirectCalls
static Int (*CI_cut_fail)() = C_cut_fail;
#define cut_fail() (*CI_cut_fail)()
#else
#define cut_fail() C_cut_fail()
#endif

/*  void cut_succeed(void) */
extern Int PROTO(C_cut_succeed,(void));
#ifdef IndirectCalls
static Int (*CI_cut_succeed)() = C_cut_succeed;
#define cut_succeed() (*CI_cut_succeed)()
#else
#define cut_succeed() C_cut_succeed()
#endif

#ifdef  SFUNC

#define SFArity  0
extern Term *ArgsOfSFTerm();
#ifdef IndirectCalls
static Term *((*CI_ArgsOfSFTerm)()) = ArgsOfSFTerm;
#define ArgsOfSFTerm(T) (*CI_ArgsOfSFTerm)(T)
#endif
extern Term MkSFTerm();
#ifdef IndirectCalls
static Term (*CI_MkSFTerm)() = MkSFTerm;
#define MkSFTerm(F,N,A,EV) (*CI_MkSFTerm)(F,N,A,EV)
#endif

#endif /* SFUNC */


#define InitCPred(N,A,F)  UserCPredicate(N,F,A)

