/*************************************************************************
*									 *
*	 YAP Prolog 	%W% %G%		*
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		clause.h						 *
* Last rev:								 *
* mods:									 *
* comments:	clause info						 *
*									 *
*************************************************************************/

#include "Yatom.h"

/* Either we are assembling clauses or indexing code */

#define ASSEMBLING_CLAUSE	0
#define ASSEMBLING_INDEX	1

/* This information is put at the start of every clause */

#define VarCl		0x0000	/* The clause's first argument is a var */
#define ListCl		0x0001	/* The clause's first argument is a list */
#define ApplCl		0x0002	/* The clause's first argument is an Appl */
#define AtCl		0x0003	/* The clause's first argument is a const */

/* If the firs argument is a list, then we care about what 
   we have in its head */
#define FHeadVar	0x0000	/* The head of the first argument is a var */
#define FHeadList	0x0004	/* The head of the first argument is a list */
#define FHeadAppl	0x0008	/* The head of the first argument ia an Appl */
#define FHeadCons	0x000c	/* The head of the first argument is a cons */

/* If the first argument is a variable, then it may be tipified later */
#define FIsVar		0x0010	/* ... :- var(X)... */
#define FIsAtom		0x0020	/* ... :- atom(X) .... */
#define FIsNum		0x0040  /* ... :- integer(X) ...
				   ... :- number(X) ... */
#define FIsPrimi	0x0080	/* ... :- atomic(X) ...
				   ... :- primitive(X) ... */

#define KindOfArg(X)	( *SmallPtr(((CODEADDR)(X))-SmallSize) & 0x03 )
#define KindOfListArg(X)  ( (*SmallPtr(((CODEADDR)(X))-SmallSize) >> 2) & 0x03 )
#define KindOfBipArg(X)  ( *SmallPtr(((CODEADDR)(X))-SmallSize) )

#define NextClause(X)	(((yamop *)X)->u.ld.d)

#define PredFirstClause		0
#define PredMiddleClause	1
#define PredLastClause		2

typedef struct {
 /* This info is used by the indexing algorithm and by the dynamic clauses.
    It is either the value of the first arg for static clauses or a pointer
    to the previous clause */
  union {
    CELL		ClValue;        /* indexable clause */
    yamop              *ClPrevious;     /* immediate update clause */
    Int                 ClOwner;        /* indexing code for logical semantics */
  } u;
  /* A set of flags describing info on the clause */
  OPREG         ClFlags;
  /* The instructions, at least one of the form sl */
  yamop         ClCode[1];
} Clause;

typedef struct {
  Clause        cl;
  /* where we are in the trail */
  Int          ClTrail;
  /* who is using this clause */
  Int          ClMark;
  /* beginning of clause */
  Clause        *my_clause;
} ExtendedClause;



#define ClauseCodeToClause(p)    ((Clause *)((CODEADDR)(p)-(CELL)(((Clause *)NULL)->ClCode)))
#define ClauseFlagsToClause(p)    ((Clause *)((CODEADDR)(p)-(CELL)(&(((Clause *)NULL)->ClFlags))))
extern int c_mask;

extern CELL c_store;

extern int pred_type;

extern PredEntry *pred_p;

extern PredEntry *CurrentPred;

