/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		load_dl.c						 *
* comments:	dl based dynamic loaderr of external routines		 *
*               tested on i486-linuxelf					 *
*************************************************************************/

#include "Yap.h"
#include "Foreign.h"

#if LOAD_DL

#include <dlfcn.h>
#include <stdio.h>

/*
 *   YAPFindExecutable(argv[0]) should be called on yap initialization to
 *   locate the executable of Yap
*/
void
YAPFindExecutable(char *name)
{
  /* not really needed for dl version */
  strcpy(YapExecutable,"./yap");
}


/*
 * LoadForeign(ofiles,libs,proc_name,init_proc) dynamically loads foreign
 * code files and libraries and locates an initialization routine
*/
Int
LoadForeign(StringList ofiles, StringList libs,
	       char *proc_name,	YapInitProc *init_proc)
{

  while (ofiles) {
    void *handle;
    char fullpath[FILENAME_MAX];

    /* dlopen wants to follow the LD_CONFIG_PATH */
    if (!TrueFileName(ofiles->s, fullpath)) {
      strcpy(LoadMsg, "[ Trying to open unexisting file in LoadForeign ]");
      return LOAD_FAILLED;
    }
#ifdef __osf__
    if((handle=dlopen(fullpath,RTLD_LAZY)) == 0)
#else
    if((handle=dlopen(fullpath,RTLD_LAZY|RTLD_GLOBAL)) == 0)
#endif
    {
fprintf(stderr,"calling dlopen with error %s\n", dlerror());
/*      strcpy(LoadMsg,dlerror());*/
      return LOAD_FAILLED;
    }

    if (!*init_proc)
      *init_proc = (YapInitProc) dlsym(handle,proc_name);

    ofiles = ofiles->next;
  }

  if(! *init_proc) {
    strcpy(LoadMsg,"Could not locate initialization routine");
    return LOAD_FAILLED;
  }

  /* load libraries first so that their symbols are available to
     other routines */
  while (libs) {
    void *handle;
    char s[FILENAME_MAX];

    if (libs->s[0] == '-') {
      strcpy(s,"lib");
      strcat(s,libs->s+2);
      strcat(s,".so");
    } else {
      strcpy(s,libs->s);
    }

#ifdef __osf__
    if((handle=dlopen(s,RTLD_LAZY)) == NULL)
#else
    if((handle=dlopen(s,RTLD_LAZY|RTLD_GLOBAL)) == NULL)
#endif
    {
      strcpy(LoadMsg,dlerror());
      return LOAD_FAILLED;
    }
    libs = libs->next;
  }

  return LOAD_SUCCEEDED;
}

#endif




