/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		consult.yap						 *
* Last rev:	8/2/88							 *
* mods:									 *
* comments:	Consulting Files in YAP					 *
*									 *
*************************************************************************/

% if source_mode is on, then the source for the predicates
% is stored with the code
source_mode(Old,New) :-
	('$get_value'('$source_mode',on) ->
			Old = on ;
			Old = off ),
	'$set_value'('$source_mode',New).

source :- source_mode(X,on).
no_source :- source_mode(X,off).

ensure_loaded(V) :- var(V), !,
	throw(error(instantiation_error,ensure_loaded(V))).
ensure_loaded([]) :- !.
ensure_loaded([F|Fs]) :- !,
	ensure_loaded(F),
	ensure_loaded(Fs).
ensure_loaded(X) :- atom(X), !,
	'$find_in_path'(X,Y),
	( open(Y,'$csult',Stream), !,
	     ( '$loaded'(Stream) ->
		(  $consulting_file_name(Stream,TFN),
		    '$recorded'($module,$module(TFN,M,P),_) ->
			$current_module(T,T), $import(P,M,T)
		;
		   true
		)
	     ;
	       '$record_loaded'(Stream),
	       '$reconsult'(X,Stream)
	     ),
	      close(Stream)
	;
		
	throw(error(permission_error(input,stream,X),ensure_loaded(X)))
	).
ensure_loaded(library(X)) :- !,
	'$find_in_path'(library(X),Y),
	( open(Y,'$csult',Stream), !,
	     ( '$loaded'(Stream) ->
		(  $consulting_file_name(Stream,TFN),
		    '$recorded'($module,$module(TFN,M,P),_) ->
			$current_module(T,T), $import(P,M,T)
		;
		   true
		)
	     ;
	       '$record_loaded'(Stream),
	       '$reconsult'(Y,Stream)
	     ),
	      close(Stream)
	;
	throw(error(permission_error(input,stream,X),ensure_loaded(library(X))))
	).
ensure_loaded(V) :- 
	throw(error(type_error(atom,V),ensure_loaded(V))).

% leave compile mode to 1 for native code.
compile(A) :- '$compile_mode'(Old,0), reconsult(A), '$compile_mode'(_,Old).

reconsult(V) :- var(V), !,
	throw(error(instantiation_error,reconsult(V))).
reconsult([]) :- !.
reconsult([F|Fs]) :- !,
	reconsult(F),
	reconsult(Fs).
reconsult(X) :- atom(X), !,
	'$find_in_path'(X,Y),
	( open(Y,'$csult',Stream), !,
		'$record_loaded'(Stream),
		'$reconsult'(X,Stream), close(Stream)
	;
		throw(error(permission_error(input,stream,Y),reconsult(X)))
	).
reconsult(library(X)) :- !,
	'$find_in_path'(library(X),Y),
	( open(Y,'$csult',Stream), !,
		'$record_loaded'(Stream),
		'$reconsult'(library(X),Stream), close(Stream)
	;
		throw(error(permission_error(input,stream,Stream),reconsult(library(X))))
	).
ensure_loaded(V) :- 
	throw(error(type_error(atom,V),reconsult(V))).

'$reconsult'(F,Stream) :-
	'$getcwd'(OldD),
	'$set_consulting_file'(Stream),
	H0 is heapused, T0 is cputime,
	current_stream(File,_,Stream),
	'$get_value'('$consulting',Old),
	'$set_value'('$consulting',false),
	'$start_reconsulting'(F),
	'$start_consult'(reconsult,File,LC),
	'$recorda'('$initialization','$',_),
	( '$get_value'('$verbose',on) ->
		tab(user_error,LC),
		'$format'(user_error, "[ reconsulting ~w... ]~n", [F])
	    ; true ),
	'$loop'(Stream,reconsult),
	'$end_consult',
	'$exec_initialization_goals',
	'$clear_reconsulting',
	( LC == 0 -> prompt(_,'   |: ') ; true),
	( '$get_value'('$verbose',on) ->
		tab(user_error,LC) ;
	true ),	
	H is heapused-H0, T is cputime-T0,
	'$format'(user_error, "[ ~w reconsulted ~w bytes in ~w seconds ]~n", [F,H,T]),
	'$set_value'('$consulting',Old),
	'$cd'(OldD),
	!.

'$start_reconsulting'(F) :-
	'$recorda'('$reconsulted','$',_),
	'$recorda'('$reconsulting',F,_).

'EMACS_FILE'(F,File0) :-
	format('''EMACS_RECONSULT''(~w).~n',[File0]),
	'$getcwd'(OldD),
	open(F,$csult,Stream),
	'$find_in_path'(File0,File),
	open(File,$csult,Stream0),
	'$set_consulting_file'(Stream0),
	H0 is heapused, T0 is cputime,
	'$get_value'('$consulting',Old),
	'$set_value'('$consulting',false),
	'$start_reconsulting'(File),
	'$start_consult'(reconsult,File,LC),
	'$recorda'('$initialization','$',_),
	( '$get_value'('$verbose',on) ->
		tab(user_error,LC),
		'$format'(user_error, "[ reconsulting ~w... ]~n", [F])
	    ; true ),
	'$loop'(Stream,reconsult),
	'$end_consult',
	'$exec_initialization_goals',
	'$clear_reconsulting',
	( LC == 0 -> prompt(_,'   |: ') ; true),
	( '$get_value'($verbose,on) ->
		tab(user_error,LC) ;
	true ),	
	H is heapused-H0, T is cputime-T0,
	'$format'(user_error, "[ ~w reconsulted ~w bytes in ~w seconds ]~n", [F,H,T]),
	'$set_value'('$consulting',Old),
	'$cd'(OldD),
	!.


initialization(G) :-
	'$recorda'('$initialization',G,_),
	fail.
initialization(_).

