/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		utils.yap						 *
* Last rev:	8/2/88							 *
* mods:									 *
* comments:	Some utility predicates available in yap		 *
*									 *
*************************************************************************/

once(G) :- '$user_call'(G), !.

op(P,T,V) :- var(P), !,
	throw(error(instantiation_error,op(P,T,V))).
op(P,T,V) :- \+integer(P), !,
	throw(error(type_error(integer,P),op(P,T,V))).
op(P,T,V) :- (P < 0 ; P > 1200), !,
	throw(error(domain_error(operator_priority,P),op(P,T,V))).
op(P,T,V) :- var(T), !,
	throw(error(instantiation_error,op(P,T,V))).
op(P,T,V) :- \+atom(T), !,
	throw(error(type_error(atom,T),op(P,T,V))).
op(P,T,V) :- var(V), !,
	throw(error(instantiation_error,op(P,T,V))).
op(P,T,V) :-
	\+ atom(V), \+ '$check_list_of_operators'(V, op(P,T,V)),
	throw(error(type_error(list,V),op(P,T,V))).
op(P,T,V) :- '$op2'(P,T,V).

'$check_list_of_operators'(V, T) :- var(V), !,
	throw(error(instantiation_error,T)).
'$check_list_of_operators'([], _).
'$check_list_of_operators'([H|L], T) :-
	'$check_if_operator'(H,T).
	'$check_list_of_operators'(L, T).

'$check_if_operator'(H,T) :- var(H), !,
	throw(error(instantiation_error,T)).
'$check_if_operator'(H,T) :- atom(H), !.
'$check_if_operator'(H,T) :-
	throw(error(type_error(atom,H),T)).

'$op2'(P,T,[]) :- !.
'$op2'(P,T,[A|L]) :- !, '$op'(P,T,A), '$op2'(P,T,L).
'$op2'(P,T,A) :- atom(A), '$op'(P,T,A).

'$op'(P,T,',') :- !,
	throw(error(permission_error(modify,operator,','),op(P,T,','))).
'$op'(P,T,A) :- '$opdec'(P,T,A).

%%% Operating System utilities

cd(A) :- atom(A), !, atom_chars(A,S), '$cd'(S).
cd(S) :- '$cd'(S).

getcwd(D) :- '$getcwd'(SD), atom_chars(D, SD).

system(A) :- atom(A), !, atom_chars(A,S), '$system'(S).
system(S) :- '$system'(S).

shell(A) :- atom(A), !, atom_chars(A,S), '$shell'(S).
shell(S) :- '$shell'(S).

rename(Old,New) :- atom(Old), atom(New), !,
	name(Old,SOld), name(New,SNew),
	'$rename'(SOld,SNew).

environ(Na,Val) :- atom(Na), !,
	'$getenv'(Na,Val).
environ(Na,Val) :-
	'$environ_enum'(0,I),
	( '$environ'(I,S) -> '$environ_split'(S,SNa,SVal) ; !, fail ),
	atom_chars(Na, SNa),
	atom_chars(Val, SVal).

'$environ_enum'(X,X).
'$environ_enum'(X,X1) :-
	Xi is X+1,
	'$environ_enum'(Xi,X1).

'$environ_split'([61|SVal], [], SVal) :- !.
'$environ_split'([C|S],[C|SNa],SVal) :-
	'$environ_split'(S,SNa,SVal).

putenv(Na,Val) :-
	'$putenv'(Na,Val).

'$environ_enum'(X,X).
'$environ_enum'(X,X1) :-
	Xi is X+1,
	'$environ_enum'(Xi,X1).

'$environ_split'([61|SVal], [], SVal) :- !.
'$environ_split'([C|S],[C|SNa],SVal) :-
	'$environ_split'(S,SNa,SVal).

unix(V) :- var(V), !,
	throw(error(instantiation_error,unix(V))).
unix(argv(L)) :- (var(L) ; atom(L)), !, '$argv'(L).
unix(argv(V)) :-
	throw(error(type_error(atomic,V),unix(argv(V)))).
unix(cd) :- cd('~').
unix(cd(V)) :- var(V), !,
	throw(error(instantiation_error,unix(cd(V)))).
unix(cd(A)) :- atomic(A), !, cd(A).
unix(cd(V)) :-
	throw(error(type_error(atomic,V),unix(cd(V)))).
unix(environ(X,Y)) :- environ(X,Y).
unix(getcwd(X)) :- getcwd(X).
unix(shell(V)) :- var(V), !,
	throw(error(instantiation_error,unix(shell(V)))).
unix(shell(A)) :- atomic(A), !, shell(A).
unix(shell(V)) :-
	throw(error(type_error(atomic,V),unix(shell(V)))).
unix(system(V)) :- var(V), !,
	throw(error(instantiation_error,unix(system(V)))).
unix(system(A)) :- atomic(A), !, system(A).
unix(system(V)) :-
	throw(error(type_error(atomic,V),unix(system(V)))).
unix(shell) :- sh.
unix(putenv(X,Y)) :- '$putenv'(X,Y).


alarm(Interval, Goal, Left) :-
	recorded('$alarm_handler',_, Ref), erase(Ref), fail.
alarm(Interval, Goal, Left) :-
	'$current_module'(M),
	'$recordz'('$alarm_handler',M:Goal,_),
	'$alarm'(Interval, Left).

%%% Saving and restoring a computation

save(A) :- atom(A), !, name(A,S), '$save'(S).
save(S) :- '$save'(S).

% not very useful at the moment
/*
save_program(A) :- atom(A), !, name(A,S), '$save_program'(S).
save_program(S) :- '$save_program'(S).
*/
save_program(S) :- save(S), abort.

restore(A) :- atom(A), !, name(A,S), '$restore'(S).
restore(S) :- '$restore'(S).

%%% Data base predicates -> the interface to the external world

recorda(Key,Term,Ref) :- '$recorda'(Key,Term,Ref).
recordz(Key,Term,Ref) :- '$recordz'(Key,Term,Ref).

recordaifnot(Key,Term,Ref) :- '$recordaifnot'(Key,Term,Ref).
recordzifnot(Key,Term,Ref) :- '$recordzifnot'(Key,Term,Ref).

%%% Atoms with value

get_value(X,Y) :- '$get_value'(X,Y).
set_value(X,Y) :- '$set_value'(X,Y).

%%% current ....

current_atom(A) :-				% check
	atom(A), !.
current_atom(A) :-				% generate
	'$current_atom'(A).

current_predicate(A,T) :- var(T), !,            % no module given
	source_module(M),
	current_predicate(A,M:T).
current_predicate(A,M:T) :-			% module specified
	(M==primitives; current_module(M)),
	call(M:'$current_predicate2'(A,T)),
	(M\=primitives -> \+ system_predicate(T); true ).

'$current_predicate2'(A,T) :-			% only for the predicate
	atom(A), !, '$pred_defined_for'(A,T),
	'$pred_exists'(T).
'$current_predicate2'(A,T) :-			% generate them all
	'$current_predicate'(A,T),
	'$pred_exists'(T).

'$system_predicate'(Pred) :-
	'$flags'(Pred,Flags,_),
	 Flags /\ 8'40000 =\= 0.

system_predicate(P) :- '$system_predicate'(P).

system_predicate(A,P) :-			% check
	nonvar(P), !,
	call(primitives:'$system_predicate'(P)),
	functor(P,A,_).

system_predicate(A,P) :-			% generate
	call(primitives:$current_predicate2(A,P)),
	'$system_predicate'(P).

%%% User interface for statistics

statistics :- 
	T is cputime, 
	'$statistics_heap_info'(HpSpa, HpInUse), 
	write(user,'Heap space : '), write(user,HpSpa), nl(user),
	tab(user,8), write(user,'Heap in use: '), write(user,HpInUse),
	'$statistics_heap_max'(HpMax), 
	write(user,', max. used: '), write(user,HpMax), nl(user),
	'$statistics_trail_info'(TrlSpa, TrlInUse), 
	write(user,'Trail space : '), write(user,TrlSpa), nl(user),
	tab(user,8), write(user,'Trail in use: '), write(user,TrlInUse),
	'$statistics_trail_max'(TrlMax), 
	( TrlMax \= TrlSpa -> write(user,', max. used: '), write(user,TrlMax) ;
			write(user,', maximum used ') ), nl(user),
	'$statistics_stacks_info'(StkSpa, GlobInU, LocInU), 
	write(user,'Stack space : '), write(user,StkSpa), nl(user),
	tab(user,8), write(user,'Global in use: '), write(user,GlobInU),
	'$statistics_global_max'(GlobMax), 
	( GlobMax \= StkSpa ->
	    write(user,', max. used: '), write(user,GlobMax) ;
	    true ),
	    nl(user),
	tab(user,8), write(user,'Local in use: '), write(user,LocInU),
	'$statistics_local_max'(LocMax), 
	( LocMax \= StkSpa -> write(user,', max. used: '), write(user,LocMax) ;
			  true ), nl(user),
	( GlobMax = StkSpa -> tab(user,8),
		write(user,'Stack space entirely used'), nl(user) ;
		true ),
	nl(user),
	nl(user),
	'$inform_heap_overflows'(NOfHO,TotHOTime),
	write(TotHOTime), write(' msec. for '),
	write(NOfHO),
	write(' heap overflows.'), nl(user),
	'$inform_stack_overflows'(NOfSO,TotSOTime),
	write(TotSOTime), write(' msec. for '),
	write(NOfSO),
	write(' stack overflows.'), nl(user),
	'$inform_trail_overflows'(NOfTO,TotTOTime),
	write(TotTOTime), write(' msec. for '),
	write(NOfTO),
	write(' trail overflows.'), nl(user),
	'$inform_gc'(NOfGC,TotGCTime,TotGCSize),
	write(TotGCTime), write(' msec. for '),
	write(NOfGC),
	write(' garbage collections which collected '),
	write(TotGCSize),write(' bytes.'), nl(user),
	write(user,'Runtime : '), write(user,T),
	'$set_value'('$last_runtime',T),
	write(user,' sec.'), nl(user).

statistics(runtime,[T,L]) :-
	'$runtime'(T,L).
statistics(cputime,[T,L]) :-
	'$cputime'(T,L).
statistics(walltime,[T,L]) :-
	'$walltime'(T,L).
%statistics(core,[_]).
%statistics(memory,[_]).
statistics(heap,[Hp,HpF]) :-
	'$statistics_heap_info'(HpM, Hp),
	HpF is HpM-Hp.
statistics(program,Info) :-
	statistics(heap,Info).
statistics(global_stack,[GlobInU,GlobFree]) :-
	'$statistics_stacks_info'(StkSpa, GlobInU, LocInU),
	GlobFree is StkSpa-GlobInU-LocInU.
statistics(local_stack,[LocInU,LocFree]) :-
	'$statistics_stacks_info'(StkSpa, GlobInU, LocInU),
	LocFree is StkSpa-GlobInU-LocInU.
statistics(trail,[TrlInUse,TrlFree]) :-
	'$statistics_trail_info'(TrlSpa, TrlInUse),
	TrlFree is TrlSpa-TrlInUse.
statistics(garbage_collection,[NOfGC,TotGCSize,TotGCTime]) :-
	'$inform_gc'(NOfGC,TotGCTime,TotGCSize).
statistics(stack_shifts,[NOfHO,NOfSO,NOfTO]) :-
	'$inform_heap_overflows'(NOfHO,_),
	'$inform_stack_overflows'(NOfSO,_),
	'$inform_trail_overflows'(NOfTO,_).


%%% The unknown predicate,
%	informs about what the user wants to be done when
%	there are no clauses for a certain predicate */

unknown(V0,V) :- var(V), !,
	'$ask_unknown_flag'(V),
	V = V0.
unknown(_,New) :- 
	'$valid_unknown_handler'(New), fail.
unknown(Old,New) :-
	'$recorded'('$unknown','$unknown'(_,MyOld),Ref), !,
	erase(Ref),
	'$cleanup_unknown_handler'(MyOld,Old),
	'$new_unknown'(New).
unknown(fail,New) :-
	'$new_unknown'(New).

'$valid_unknown_handler'(V) :-
	var(V), !,
	throw(error(instantiation_error,yap_flag(unknown,S))).
'$valid_unknown_handler'(fail) :- !.
'$valid_unknown_handler'(error) :- !.
'$valid_unknown_handler'(warning) :- !.
'$valid_unknown_handler'(S) :-
	functor(S,_,1),
	arg(1,S,A),
	var(A), 
	\+ '$undefined'(S),
	!.
'$valid_unknown_handler'(S) :-
	throw(error(domain_error(flag_value,unknown+S),yap_flag(unknown,S))).

'$ask_unknown_flag'(Old) :-
	'$recorded'('$unknown','$unkonwn'(_,MyOld),Ref), !,
	'$cleanup_unknwon_handler'(MyOld,Old).
'$ask_unknown_flag'(fail).

'$cleanup_unknown_handler'('$unknown_error'(_),error) :- !.
'$cleanup_unknown_handler'('$unknown_warning'(_),warning) :- !.
'$cleanup_unknown_handler'(Handler, Handler).

'$new_unknown'(fail) :- !.
'$new_unknown'(error) :- !,
	'$recorda'('$unknown','$unknown'(P,'$unknown_error'(P)),_).
'$new_unknown'(warning) :- !,
	'$recorda'('$unknown','$unknown'(P,'$unknown_warning'(P)),_).
'$new_unknown'(X) :-
	arg(1,X,A),
	'$recorda'('$unknown','$unknown'(A,X),_).

'$unknown_error'(P) :-
	throw(error(unknown,P)).

'$unknown_warning'(P) :-
	P=M:F,
	functor(F,Na,Ar),
	format(user_error,"[ EXISTENCE ERROR: ~w, procedure ~w:~w/~w undefined ]~n",
	[P,M,Na,Ar]),
	fail.

predicate_property(Mod:Pred,Prop) :- !,
	call(Mod:'$predicate_property2'(Pred,Prop)).
predicate_property(Pred,Prop) :- 
	'$predicate_property2'(Pred,Prop).

'$predicate_property2'(Pred,Prop) :- var(Pred), !,
	'$current_predicate'(_,Pred),
	'$pred_exists'(Pred),
	'$predicate_property'(Pred,Prop).
'$predicate_property2'(Pred,Prop) :- 
	'$predicate_property'(Pred,Prop),
	'$pred_exists'(Pred).

'$predicate_property'(P,built_in) :- 
	'$system_predicate'(P), !.
'$predicate_property'(P,dynamic) :-
	'$is_dynamic'(P).
'$predicate_property'(P,compiled) :-
	\+ '$is_dynamic'(P).
'$predicate_property'(P,fast) :-
	\+ '$is_dynamic'(P),
	'$flags'(P,F,_F),
	F /\ 16'20000 =\= 0.
'$predicate_property'(P,multifile) :-
	functor(P,N,A),
	( '$recorded'('$multifile','$multifile'(N,A,_),_) -> true ).
'$predicate_property'(P,imported_from(Mod)) :-
	functor(P,N,A),
	'$recorded'($module,$module(_TFN,Mod,Publics),_),
	$member(N/A,Publics).	/* defined in modules.yap */
'$predicate_property'(P,exported) :-
	functor(P,N,A),
	$current_module(M),
	'$recorded'($module,$module(_TFN,M,Publics),_),
	$member(N/A,Publics).	/* defined in modules.yap */

%%% Some "dirty" predicates

yap_flag(V,Out) :-
	var(V), !,
	'$show_yap_flag_opts'(V,Out).

% do or do not indexation
yap_flag(index,X) :- var(X), !,
	 ( '$get_value'('$doindex',true) -> X=on ; X=off).
yap_flag(index,on)  :- !, '$set_value'('$doindex',true).
yap_flag(index,off) :- !, '$set_value'('$doindex',[]).

% do or do not machine code
yap_flag(fast,on) :- '$set_value'('$fast',true).
yap_flag(fast,off) :- !, '$set_value'('$fast',[]).

% hide/unhide atoms
yap_flag(hide,Atom) :- !, hide(Atom).
yap_flag(unhide,Atom) :- !, unhide(Atom).

% control garbage collection
yap_flag(gc,V) :-
	var(V), !,
	( '$get_value'('$gc',[]) -> V = off ; V = on).
yap_flag(gc,on) :- '$set_value'('$gc',true).
yap_flag(gc,off) :- '$set_value'('$gc',[]).
yap_flag(gc_margin,N) :- 
	var(N) -> 
		 '$get_value'('$gc_margin',N)
        ;
        integer(N) ->
	         '$set_value'('$gc_margin',N).
yap_flag(gc_trace,V) :-
	var(V), !,
	'$get_value'('$gc_trace',N1),
	'$get_value'('$gc_verbose',N2),
	'$yap_flag_show_tracing'(N1, N2, V).
yap_flag(gc_trace,on) :-
	'$set_value'('$gc_trace',true),
	'$set_value'('$gc_verbose',[]).
yap_flag(gc_trace,verbose) :-
	'$set_value'('$gc_trace',[]),
	'$set_value'('$gc_verbose',true).
yap_flag(gc_trace,off) :-
	'$set_value'('$gc_trace',[]),
	'$set_value'('$gc_verbose',[]).
yap_flag(syntax_errors, V) :- var(V), !,
	'$get_read_error_handler'(V).
yap_flag(syntax_errors, Option) :-
	'$set_read_error_handler'(Option).
% compatibility flag
yap_flag(enhanced,on) :- '$set_value'('$enhanced',true).
yap_flag(enhanced,off) :- '$set_value'('$enhanced',[]).
%
% show state of $
%
yap_flag(dollar_as_lower_case,V) :-
	var(V), !,
	'$type_of_char'(36,T),
	(T = 3 -> V = on ; V = off).
%
% make $a a legit atom
%
yap_flag(dollar_as_lower_case,on) :-
	'$change_type_of_char'(36,3).
%
% force quoting of '$a'
%
yap_flag(dollar_as_lower_case,off) :- 
	'$change_type_of_char'(36,7).

yap_flag(profiling,X) :- (var(X); X = on; X = off), !,
	'$is_profiled'(X).

yap_flag(bounded,X) :-
	var(X), !,
	'$access_yap_flags'(0, X1),
	'$transl_to_true_false'(X1,X).
yap_flag(bounded,X) :-
	(X = true ; X = false), !,
	throw(error(permission_error(modify,flag,bounded),yap_flag(bounded,X))).
yap_flag(bounded,X) :-
	throw(error(domain_error(flag_value,bounded+X),yap_flag(bounded,X))).

yap_flag(max_arity,X) :-
	var(X), !,
	'$access_yap_flags'(1, X1),
	'$transl_to_arity'(X1,X).
yap_flag(max_arity,X) :-
	integer(X), X > 0, !,
	throw(error(permission_error(modify,flag,max_arity),yap_flag(max_arity,X))).
yap_flag(max_arity,X) :-
	throw(error(domain_error(flag_value,max_arity+X),yap_flag(max_arity,X))).

yap_flag(integer_rounding_function,X) :-
	var(X), !,
	'$access_yap_flags'(2, X1),
	'$transl_to_rounding_function'(X1,X).
yap_flag(integer_rounding_function,X) :-
	(X = down; X = toward_zero), !,
	throw(error(permission_error(modify,flag,integer_rounding_function),yap_flag(integer_rounding_function,X))).
yap_flag(integer_rounding_function,X) :-
	throw(error(domain_error(flag_value,integer_rounding_function+X),yap_flag(integer_rounding_function,X))).

yap_flag(max_integer,X) :-
	var(X), !,
	'$access_yap_flags'(3, X).
yap_flag(max_integer,X) :-
	integer(X), X > 0, !,
	throw(error(permission_error(modify,flag,max_integer),yap_flag(max_integer,X))).
yap_flag(max_integer,X) :-
	throw(error(domain_error(flag_value,max_integer+X),yap_flag(max_integer,X))).

yap_flag(min_integer,X) :-
	var(X), !,
	'$access_yap_flags'(4, X).
yap_flag(min_integer,X) :-
	integer(X), X < 0, !,
	throw(error(permission_error(modify,flag,min_integer),yap_flag(min_integer,X))).
yap_flag(min_integer,X) :-
	throw(error(domain_error(flag_value,min_integer+X),yap_flag(min_integer,X))).

yap_flag(char_conversion,X) :-
	var(X), !,
	'$access_yap_flags'(5, X1),
	'$transl_to_on_off'(X1,X).
yap_flag(char_conversion,X) :-
	'$transl_to_on_off'(X1,X), !,
	( X1 = 0 ->
	    throw(error(permission_error(modify,flag,char_conversion),yap_flag(char_conversion,X)));
	    '$set_yap_flags'(5,X1)
	).
yap_flag(char_conversion,X) :-
	throw(error(domain_error(flag_value,char_conversion+X),yap_flag(char_conversion,X))).

yap_flag(double_quotes,X) :-
	var(X), !,
	'$access_yap_flags'(6, X1),
	'$transl_to_trl_types'(X1,X).
yap_flag(double_quotes,X) :-
	'$transl_to_trl_types'(X1,X), !,
	'$set_yap_flags'(6,X1).
yap_flag(double_quotes,X) :-
	throw(error(domain_error(flag_value,double_quotes+X),yap_flag(double_quotes,X))).

yap_flag(debug,X) :-
	var(X), !,
	('$get_value'(debug,1) ->
	    X = on
	;
	    X = off
	).
yap_flag(debug,X) :-
	'$transl_to_on_off'(_,X), !,
	(X = on -> debug ; nodebug).
yap_flag(debug,X) :-
	throw(error(domain_error(flag_value,debug+X),yap_flag(debug,X))).

yap_flag(unknown,X) :-
	var(X), !,
	unknown(X,_).
yap_flag(unknown,N) :-
	unknown(_,N).

'$transl_to_true_false'(0,false).
'$transl_to_true_false'(1,true).

'$transl_to_on_off'(0,on).
'$transl_to_on_off'(1,off).

'$transl_to_arity'(X1,X) :- X1 < 0, !, X = unbounded.
'$transl_to_arity'(X,X).

'$transl_to_rounding_function'(0,down).
'$transl_to_rounding_function'(1,toward_zero).

'$transl_to_trl_types'(0,chars).
'$transl_to_trl_types'(1,codes).
'$transl_to_trl_types'(2,atom).

'$yap_flag_show_tracing'(true, _, on) :- !.
'$yap_flag_show_tracing'(_, true, verbose) :- !.
'$yap_flag_show_tracing'(_, _, off).

'$show_yap_flag_opts'(V,Out) :-
	(
	    V = bounded ;
	    V = char_conversion ;
	    V = debug ;
	    V = dollar_as_lower_case ;
	    V = double_quotes ;
%	    V = fast  ;
	    V = gc    ;
	    V = gc_margin    ;
	    V = gc_trace     ;
%	    V = hide  ;
	    V = index ;
	    V = integer_rounding_function ;
	    V = max_arity ;
	    V = max_integer ;
	    V = min_integer ;
	    V = profiling ;
	    V = syntax_errors ;
	    V = unknown
	),
	yap_flag(V, Out).

current_prolog_flag(V,Out) :-
	(var(V) ; atom(V) ), !,
	'$show_yap_flag_opts'(V,NOut),
	NOut = Out.
current_prolog_flag(V,Out) :-
	throw(error(type_error(atom,V),current_prolog_flag(V,Out))).

set_prolog_flag(F,V) :-
	var(F), !,
	throw(error(instantiation_error,set_prolog_flag(F,V))).
set_prolog_flag(F,V) :-
	var(V), !,
	throw(error(instantiation_error,set_prolog_flag(F,V))).
set_prolog_flag(F,V) :-
	\+ atom(F), !,
	throw(error(type_error(atom,F),set_prolog_flag(F,V))).
set_prolog_flag(F,V) :-
	yap_flag(F,V).

% Only efective if yap compiled with -DDEBUG
% this predicate shows the code produced by the compiler
'$show_code' :- '$debug'(0'f).

'$pred_exists'(Pred) :- '$is_dynamic'(Pred), !.
'$pred_exists'(Pred) :- \+ '$undefined'(Pred).


grow_heap(X) :- $grow_heap(X).
grow_stack(X) :- $grow_stack(X).

%
% gc() expects to be called from "call". Make sure it has an
% environment to return to.
%
%garbage_collect :- save(dump), '$gc',  save(dump2).
garbage_collect :- '$gc'.
gc :- yap_flag(gc,on).
nogc :- yap_flag(gc,off).

'$force_environment_for_gc'.

profile_data(P, Parm, Data) :- var(P), !,
	'$profile_data_for_var'(P, Parm, Data).
profile_data(Na/Ar, Parm, Data) :-
	'$profile_info'(Na, Ar, Stats),
	functor(P0, Na, Ar),
	'$profile_say'(Stats, Parm, Data).

'$profile_data_for_var'(P, Parm, Data) :-
	'$current_predicate'(_,P0),
	functor(P0, Name, Arity),
	P = Name/Arity,
	'$profile_info'(Name, Arity, Stats),
	'$profile_say'(Stats, Parm, Data).



'$profile_say'('$profile'(Entries, _, _), calls, Entries).
'$profile_say'('$profile'(_, _, Backtracks), retries, Backtracks).

profile_reset :-
	current_predicate(_,P0),
	functor(P0, Name, Arity),
	'$profile_reset'(Name, Arity),
	fail.
profile_reset.

atom_codes(A,L) :- nonvar(L),
	( '$good_list_of_character_codes'(L) ->
	    fail
	;
	    throw(error(representation_error(character_code_list,L),atom_codes(A,L)))
        ).
atom_codes(A,L) :- atom_chars(A,L).

'$good_list_of_character_codes'(V) :- var(V), !.
'$good_list_of_character_codes'([]).
'$good_list_of_character_codes'([X|L]) :-
	'$good_character_code'(X),
	'$good_list_of_character_codes'(L).

'$good_character_code'(X) :- var(X), !.
'$good_character_code'(X) :- integer(X), X > -2, X < 256.

atom_concat(X,Y,At) :-
	atom(At), !,
	atom_length(At,Len),
	'$atom_contact_split'(At,0,Len,X,Y).
/* Let atom_chars do our error handling */
atom_concat(X,Y,At) :-
	atom_chars(X,S1),
	atom_chars(Y,S2),
	'$append'(S1,S2,S),
	atom_chars(At,S).



'$atom_contact_split'(At,Len,Len,X,Y) :- !,
	'$atom_split'(At,Len,X,Y).
'$atom_contact_split'(At,Len1,Len,X,Y) :-
	'$atom_split'(At,Len1,X,Y).
'$atom_contact_split'(At,Len1,Len,X,Y) :-
	Len2 is Len1+1,
	'$atom_contact_split'(At,Len2,Len,X,Y).

number_codes(A,L) :- nonvar(L),
	( '$good_list_of_character_codes'(L) ->
	    fail
	;
	    throw(error(domain_error(character_code_list,L),number_codes(A,L)))
        ).
number_codes(A,L) :- number_chars(A,L).


sub_atom(At, Bef, Size, After, SubAt) :-
	var(At), !,
	throw(error(instantiation_error,sub_atom(At, Bef, Size,After, SubAt))).
sub_atom(At, Bef, Size, After, SubAt) :-
	\+ atom(At), !,
	throw(error(type_error(atom,At),sub_atom(At, Bef, Size,After, SubAt))).
sub_atom(At, Bef, Size, After, SubAt) :-
	nonvar(SubAt), \+ atom(SubAt), !,
	throw(error(type_error(atom,SubAt),sub_atom(At, Bef, Size,After, SubAt))).
sub_atom(At, Bef, Size, After, SubAt) :-
	'$check_type_sub_atom'(Bef, sub_atom(At, Bef, Size,After, SubAt)),
	'$check_type_sub_atom'(Size, sub_atom(At, Bef, Size,After, SubAt)),
	'$check_type_sub_atom'(After, sub_atom(At, Bef, Size,After, SubAt)),
	atom_codes(At, Atl),
	'$$_length1'(Atl,0,Len),
	'$split_len_in_parts'(Atl, Len, Bef, Size, After, SubAtl),
	atom_codes(SubAt, SubAtl).

'$check_type_sub_atom'(I, P) :-
	var(I), !.
'$check_type_sub_atom'(I, P) :-
	integer(I), I < 0, !,
	throw(error(domain_error(not_less_than_zero,I),P)).
'$check_type_sub_atom'(I, P) :-
	\+ integer(I), !,
	throw(error(type_error(integer,I),P)).
'$check_type_sub_atom'(I, P).

'$split_len_in_parts'(Atl, Len, Bef, Size, After, SubAt) :-
	var(Bef), !,
	'$range_var'(0, Len, Bef),
	'$split_len_in_parts2'(Atl, Len, Bef, Size, After, SubAt).
'$split_len_in_parts'(Atl, Len, Bef, Size, After, SubAt) :-
	'$split_len_in_parts2'(Atl, Len, Bef, Size, After, SubAt).

'$split_len_in_parts2'(Atl, Len, 0, Size, After, SubAt) :- !,
	'$split_len_in_parts3'(Atl, Len, Size, After, SubAt).
'$split_len_in_parts2'([_|Atl], Len0, Bef0, Size, After, SubAt) :-
	Bef is Bef0-1 ,
	Len is Len0-1,
	'$split_len_in_parts2'(Atl, Len, Bef, Size, After, SubAt).
	
	
'$split_len_in_parts3'(Atl, Len, Size, After, SubAt) :-
	var(Size), !,
	'$range_var'(0, Len, Size),
	'$split_len_in_parts4'(Atl, Len, Size, After, SubAt).
'$split_len_in_parts3'(Atl, Len, Size, After, SubAt) :-
	'$split_len_in_parts4'(Atl, Len, Size, After, SubAt).

'$split_len_in_parts4'(_, Len, 0, After, SubAt) :- !,
	After = Len,
	SubAt = [].
'$split_len_in_parts4'([Code|Atl], Len0, Size0, After, [Code|SubAt]) :-
	Size is Size0-1,
	Len is Len0-1,
	'$split_len_in_parts4'(Atl, Len, Size, After, SubAt).
	
'$range_var'(X,X,S) :- !, S = X.
'$range_var'(X1,X2,X1).
'$range_var'(X1,X2,XF) :-
	X11 is X1+1,
	'$range_var'(X11,X2,XF).


current_predicate(V) :- var(V), !,
	current_predicate(_,S),
	functor(S,Na,Ar),
	V = Na/Ar.
current_predicate(Na/Ar) :- !,
	current_predicate(Na,S),
	functor(S,Na,Ar).
current_predicate(T) :-
	throw(error(type_error(predicate_indicator,T),current_predicate(T))).


'$singletons_in_term'(T,VL) :-
	'$variables_in_term'(T,[],V10),
	'$sort'(V10, V1),
	'$non_singletons_in_term'(T,[],V20),
	'$sort'(V20, V2),	
	'$subtract_lists_of_variables'(V2,V1,VL).

'$subtract_lists_of_variables'([],VL,VL).
'$subtract_lists_of_variables'([_|_],[],[]) :- !.
'$subtract_lists_of_variables'([V1|VL1],[V2|VL2],VL) :-
	V1 == V2, !,
	'$subtract_lists_of_variables'(VL1,VL2,VL).
'$subtract_lists_of_variables'([V1|VL1],[V2|VL2],[V2|VL]) :-
	'$subtract_lists_of_variables'([V1|VL1],VL2,VL).
	
